"""
Main control point for the viral annotation process
"""
from datetime import datetime
from os import path, mkdir
import re
import warnings
import logging
import pandas as pd
import numpy as np
from skbio.io import read as read_sequence
from skbio.io import write as write_sequence

from mag_annotator.database_handler import DatabaseHandler
from mag_annotator.annotate_bins import annotate_fastas, get_fasta_name, perform_fasta_checks
from mag_annotator.utils import setup_logger
from mag_annotator.summarize_genomes import get_ids_from_annotations_all

VMAG_DBS_TO_ANNOTATE = ('kegg', 'kofam_hmm', 'kofam_ko_list', 'uniref', 'peptidase', 'pfam', 'dbcan', 'viral', 'vogdb')
VIRSORTER_COLUMN_NAMES = ['gene_name', 'start_position', 'end_position', 'length', 'strandedness',
                          'viral_protein_cluster_hit', 'viral_protein_cluster_hit_score',
                          'viral_protein_cluster_hit_evalue', 'viral_protein_cluster_category', 'pfam_hit',
                          'pfam_hit_score', 'pfam_hit_evalue', 'name']
VIRSORTER_HALLMARK_GENE_CATEGORIES = {'0', '3'}
VIRSORTER_VIRAL_LIKE_GENE_CATEGORIES = {'1', '4'}
TRANSPOSON_PFAMS = {'PF01609', 'PF00872', 'PF01610', 'PF01527', 'PF02371', 'PF01710', 'PF01385', 'PF01548', 'PF01526',
                    'PF01797', 'PF02899', 'PF05717', 'PF07592', 'PF03050', 'PF04754', 'PF04986', 'PF03400'}
CELL_ENTRY_CAZYS = {'CBM50', 'GH102', 'GH103', 'GH104', 'GH108', 'GH18', 'GH19', 'GH22', 'GH23', 'GH24', 'GH25', 'GH73',
                    'PL9', 'CBM12', 'CBM14', 'CBM18', 'CBM19', 'AA15', 'AA10', 'GH32', 'GH58', 'PL16'}
VIRAL_PEPTIDASES_MEROPS = {'A02H', 'A02G', 'A02F', 'A02E', 'A02D', 'A02C', 'A02B', 'A02A', 'A03B', 'A03A', 'A11B',
                           'A11A', 'A22B', 'A22A', 'A33', 'C01B', 'C01A', 'C02B', 'C02A', 'C03H', 'C03G', 'C03F',
                           'C03E', 'C03D', 'C03C', 'C03B', 'C03A', 'C04', 'C05', 'C06', 'C07', 'C08', 'C09', 'C104',
                           'C105', 'C107', 'C108', 'C113', 'C14B', 'C14A', 'C16B', 'C16A', 'C18', 'C19', 'C21', 'C23',
                           'C24', 'C26', 'C27', 'C28', 'C30', 'C31', 'C32', 'C33', 'C36', 'C37', 'C39', 'C40', 'C42',
                           'C44', 'C46', 'C48', 'C49', 'C51', 'C53', 'C57', 'C59', 'C60B', 'C60A', 'C62', 'C63', 'C71',
                           'C74', 'C76', 'C85B', 'C85A', 'C87', 'C89', 'C97', 'C99', 'G02', 'I02', 'I04', 'I08', 'I24',
                           'I25C', 'I25B', 'I25A', 'I29', 'I32', 'I36', 'I43', 'I50B', 'I50A', 'I51', 'I63', 'I75',
                           'I87', 'I91', 'M03C', 'M03B', 'M03A', 'M10C', 'M10B', 'M10A', 'M12C', 'M12B', 'M12A', 'M13',
                           'M14C', 'M14D', 'M14B', 'M14A', 'M15D', 'M15C', 'M15B', 'M15A', 'M16C', 'M16B', 'M16A',
                           'M20E', 'M20A', 'M20D', 'M20F', 'M20B', 'M20C', 'M23B', 'M23A', 'M27', 'M38', 'M41', 'M42',
                           'M43B', 'M43A', 'M44', 'M48C', 'M48B', 'M48A', 'M56', 'M60', 'M67C', 'M67B', 'M67A', 'M78',
                           'M79', 'M86', 'N01', 'N02', 'N04', 'N05', 'N07', 'N08', 'N09', 'N10E', 'N10D', 'N10C',
                           'N10B', 'N11', 'S01F', 'S01E', 'S01D', 'S01C', 'S01B', 'S01A', 'S03', 'S06', 'S07', 'S08C',
                           'S08B', 'S08A', 'S09D', 'S09C', 'S09B', 'S09A', 'S11', 'S12', 'S14', 'S16', 'S21', 'S24',
                           'S26C', 'S26B', 'S26A', 'S28', 'S29', 'S30', 'S31', 'S32', 'S33', 'S49C', 'S49B', 'S49A',
                           'S50', 'S53', 'S54', 'S62', 'S65', 'S69', 'S73', 'S74', 'S75', 'S77', 'S78', 'S80', 'S81',
                           'T01B', 'T01A', 'T03', 'U32', 'U40'}


def is_affi_tab_not_fasta(input_file: str) -> bool:
    """
    Checks that a file is a VIRSorter affi tab file.

    :param input_file:
        the input file that the user provided.
    :returns:
        True if there are 11 delimiters on the second line.
    """
    with open(input_file) as input_io:
        num_delimiter = input_io.readlines()[1].count('|')
    if num_delimiter == 11:
        return True
    return False


def remove_bad_chars_fasta(fasta):
    if is_affi_tab_not_fasta(fasta):
        raise ValueError("The input file format matches virsorter affi contigs not fasta, please check "
                         "that the flags match the file type: '-v' for virsorter, and '-i' for fasta.")

    new_headers = list()
    new_seqs = list()
    for seq in read_sequence(fasta, format='fasta'):
        new_header = seq.metadata['id']
        new_header = new_header.replace(';', '_').replace('=', '_')
        if new_header in new_headers:
            raise ValueError('Replacement of ; or = with _ generated redundant sequence headers, must be modified '
                             'manually')
        seq.metadata['id'] = new_header
        new_seqs.append(seq)
    return new_seqs


def remove_bad_chars_virsorter_affi_contigs(virsorter_in: str) -> str:
    """
    Determine and call the appropriate function to cleans incompatible characters from VIRSorter
    or VIRSorter2 affi headers.

    :param virsorter_in:

        Path to a affi tab file provide for DRAM by VIRSorter or VIRSorter2.
    :returns:
        String of file contents with headers cleaned.
    :raises ValueError:
        If the input format does not match VIRSorter or VIRSorter2.
    """
    if not is_affi_tab_not_fasta(virsorter_in):
        raise ValueError("The input file format does not match virsorter affi contigs, please check "
                         "that the flags match the file type: '-v' for virsorter, and '-i' for "
                         "fasta.")
    new_lines = list()
    for line in open(virsorter_in).readlines():
        line = line.strip()
        if line.startswith('>'):
            line = line.lstrip('>')
            split_line = line.split('|')
            split_line[0] = split_line[0].replace(';', '_').replace('=', '_')
            new_lines.append('>%s' % '|'.join(split_line))
        else:
            split_line = line.split('|')
            split_line[0] = split_line[0].replace(';', '_').replace('=', '_')
            new_lines.append('|'.join(split_line))
    return '%s\n' % '\n'.join(new_lines)


def remove_bad_chars(input_fasta: str = None, input_virsorter_affi_contigs: str = None, output: str = None):
    """
    Determine and call the function to cleans incompatible characters from headers of virsorter
    affi or fasta file.

    :param input_fasta:
        Path to a fasta file provide for DRAM by VIRSorter or VIRSorter2
    :param input_virsorter_affi_contigs:
        Path to a affi tab file provide for DRAM by VIRSorter
        or VIRSorter2
    :param output:
        Desired path for the new file clean of incompatible characters
    :raises ValueError:
        Check format, and invalid arguments.
    """
    if input_fasta is None and input_virsorter_affi_contigs is None:
        raise ValueError('Must provide either input fasta or virsorter affi contigs')
    elif input_fasta is not None:
        if ';' in output or '=' in output:
            raise ValueError('Cannot have = or ; in output fasta to work with DRAM-v, please change output fasta name')
        new_seqs = remove_bad_chars_fasta(input_fasta)
        write_sequence((seq for seq in new_seqs), format='fasta', into=output)
    else:
        virsorter_out = remove_bad_chars_virsorter_affi_contigs(input_virsorter_affi_contigs)
        open(output, 'w').write(virsorter_out)


def get_virsorter_hits(virsorter_affi_contigs):
    raw_file = open(virsorter_affi_contigs).read()

    virsorter_rows = list()
    for entry in raw_file[1:].split('\n>'):
        split_entry = entry.strip().split('\n')
        entry_data = split_entry[0].split('|')
        entry_name = entry_data[0]
        if '=' in entry_name or ';' in entry_name:
            raise ValueError('FASTA headers must not have = or ; before the first space (%s). To run DRAM-v you must '
                             'rerun VIRSorter with = and ; removed from the headers or run DRAM-v.py '
                             'remove_bad_characters and then rerun DRAM-v' % entry_name)
        entry_genes = split_entry[1:]
        entry_rows = [i.split('|') + [entry_name] for i in entry_genes]
        virsorter_rows += entry_rows
    virsorter_hits = pd.DataFrame(virsorter_rows, columns=VIRSORTER_COLUMN_NAMES).set_index('gene_name')
    virsorter_hits.index = [i.replace('(', '_').replace(')', '_') for i in virsorter_hits.index]
    virsorter_hits['name'] = [i.replace('(', '_').replace(')', '_') for i in virsorter_hits['name']]
    return virsorter_hits


def get_overlap(row1, row2):
    overlap_start = np.max((row1['start_position'], row2['start_position']))
    overlap_end = np.min((row1['end_position'], row2['end_position']))
    overlap_len = overlap_end - overlap_start
    if overlap_len <= 0:
        return 0, 0
    else:
        return overlap_len / (row1['end_position'] - row1['start_position']), \
               overlap_len / (row2['end_position'] - row2['start_position'])


def is_transposon(pfam_hits):
    if pd.isna(pfam_hits):
        return False
    else:
        pfams = {i[1:-1].split('.')[0] for i in re.findall(r'\[PF\d\d\d\d\d.\d*\]', pfam_hits)}
        return len(pfams & TRANSPOSON_PFAMS) > 0


def get_gene_order(dram_genes, virsorter_genes, min_overlap=.70):
    dram_genes['start_position'] = dram_genes['start_position'].astype(int)
    dram_genes['end_position'] = dram_genes['end_position'].astype(int)
    dram_genes = dram_genes.sort_values('start_position')
    dram_gene_number = 0

    virsorter_genes['start_position'] = virsorter_genes['start_position'].astype(int)
    virsorter_genes['end_position'] = virsorter_genes['end_position'].astype(int)
    virsorter_genes = virsorter_genes.sort_values('start_position')
    virsorter_gene_number = 0

    merged_genes_rows = list()
    while (dram_gene_number < dram_genes.shape[0]) and (virsorter_gene_number < virsorter_genes.shape[0]):
        dram_row = dram_genes.iloc[dram_gene_number]
        dram_gene = dram_genes.index[dram_gene_number]
        virsorter_row = virsorter_genes.iloc[virsorter_gene_number]
        virsorter_gene = virsorter_genes.index[virsorter_gene_number]
        virsorter_gene_category = virsorter_row['viral_protein_cluster_category']
        # end if we are at the end of either list
        forward_overlap, reverse_overlap = get_overlap(dram_row, virsorter_row)
        if (forward_overlap > min_overlap) and (reverse_overlap > min_overlap):
            merged_genes_rows.append((dram_gene, virsorter_gene, virsorter_gene_category))
            dram_gene_number += 1
            virsorter_gene_number += 1
        else:  # if not significantly overlap then determine first gene
            if dram_row['start_position'] < virsorter_row['start_position']:
                merged_genes_rows.append((dram_gene, None, None))
                dram_gene_number += 1
            elif dram_row['start_position'] > virsorter_row['start_position']:
                merged_genes_rows.append((None, virsorter_gene, virsorter_gene_category))
                virsorter_gene_number += 1
            else:  # if the same start positions then shorter gene is the first
                if dram_row['end_position'] < virsorter_row['end_position']:
                    merged_genes_rows.append((dram_gene, None, None))
                    dram_gene_number += 1
                elif dram_row['end_position'] > virsorter_row['end_position']:
                    merged_genes_rows.append((None, virsorter_gene,
                                              virsorter_gene_category))
                    virsorter_gene_number += 1
                else:  # genes have the same start and end position so this shouldn't happen
                    raise ValueError('This should never happen: %s, %s, %s, %s' % (dram_row.start_position,
                                                                                   dram_row.end_position,
                                                                                   virsorter_row.start_position,
                                                                                   virsorter_row.end_position))
    # clean up and add extras
    # if at end of both then just end
    if (dram_gene_number == dram_genes.shape[0]) and (virsorter_gene_number == virsorter_genes.shape[0]):
        pass
    # if not at end of dram annotations then add the rest
    elif dram_gene_number != dram_genes.shape[0]:
        for i in range(dram_gene_number, dram_genes.shape[0]):
            dram_gene = dram_genes.index[i]
            merged_genes_rows.append((dram_gene, None, None))
    # if not at end of virsorter genes then add the rest
    elif virsorter_gene_number != virsorter_genes.shape[0]:
        for i in range(virsorter_gene_number, virsorter_genes.shape[0]):
            virsorter_row = virsorter_genes.iloc[i]
            merged_genes_rows.append((None, virsorter_genes.index[i],
                                      virsorter_row['viral_protein_cluster_category']))
    # how can we be not at the end of either?
    else:
        return ValueError('How is this even possible?')
    return merged_genes_rows


def calculate_auxiliary_scores(gene_order):
    # right now saying that flanked means between here and the end of the contig
    gene_auxiliary_score_dict = dict()
    for i, (dram_gene, virsorter_gene, virsorter_category) in enumerate(gene_order):
        if dram_gene is not None:
            left_categories = [left_virsorter_category for left_dram_gene, left_viral_gene, left_virsorter_category
                               in gene_order[:i] if left_viral_gene is not None]
            hallmark_left = len(set(left_categories) & set(VIRSORTER_HALLMARK_GENE_CATEGORIES)) > 0
            viral_like_left = len(set(left_categories) & set(VIRSORTER_VIRAL_LIKE_GENE_CATEGORIES)) > 0

            right_categories = [right_virsorter_category for right_dram_gene, right_viral_gene, right_virsorter_category
                                in gene_order[i+1:] if right_viral_gene is not None]
            hallmark_right = len(set(right_categories) & set(VIRSORTER_HALLMARK_GENE_CATEGORIES)) > 0
            viral_like_right = len(set(right_categories) & set(VIRSORTER_VIRAL_LIKE_GENE_CATEGORIES)) > 0
            # if end of contig then 5:
            if i == 0 or i == (len(gene_order)-1):
                auxiliary_score = 5
            elif hallmark_left and hallmark_right:  # hallmark on both sides then cat 1
                auxiliary_score = 1
            # hallmark on one side and viral like on other then cat 2
            elif (hallmark_left and viral_like_right) or (viral_like_left and hallmark_right):
                auxiliary_score = 2
            # viral like on both side then cat 3
            elif viral_like_left and viral_like_right:
                auxiliary_score = 3
            # not end of contig and hallmark or viral like on other side
            elif (hallmark_left or viral_like_left) or (hallmark_right or viral_like_right):
                auxiliary_score = 4
            # if gene is the only virsorter viral like or hallmark on contig then make it a 4
            elif (virsorter_category in VIRSORTER_HALLMARK_GENE_CATEGORIES) or \
                 (virsorter_category in VIRSORTER_VIRAL_LIKE_GENE_CATEGORIES):
                auxiliary_score = 4
            else:  # if no viral like or hallmark genes then score is 5
                auxiliary_score = 5
            gene_auxiliary_score_dict[dram_gene] = auxiliary_score
    return gene_auxiliary_score_dict


def get_metabolic_flags(annotations, metabolic_genes, amgs, verified_amgs, scaffold_length_dict,
                        logger, length_from_end=5000):
    flag_dict = dict()
    metabolic_genes = set(metabolic_genes)
    for scaffold, scaffold_annotations in annotations.groupby('scaffold'):
        for gene, row in scaffold_annotations.iterrows():
            # set up
            flags = ''
            gene_annotations = set(get_ids_from_annotations_all(pd.DataFrame(row).transpose()).keys())  # TODO: Fix
            # is viral
            if 'vogdb_categories' in row.index and not pd.isna(row['vogdb_categories']):
                if len({'Xr', 'Xs'} & set(row['vogdb_categories'].split(';'))) > 0:
                    flags += 'V'
            # is metabolic
            if len(metabolic_genes & gene_annotations) > 0:
                flags += 'M'
            # is this a reported AMG reported
            if len(gene_annotations & set(amgs)) > 0:
                if 'M' not in flags:
                    flags += 'M'
                flags += 'K'
            # is this a experimentally verified amg
            if len(gene_annotations & set(verified_amgs)) > 0:
                flags += 'E'
            # is this gene a normal viral cell host entry gene
            if len(gene_annotations & CELL_ENTRY_CAZYS) > 0:
                flags += 'A'
            # is gene a normal virus peptidase
            if len(gene_annotations & VIRAL_PEPTIDASES_MEROPS) > 0:
                flags += 'P'
            # if there is a transposon in the contig
            if scaffold_annotations['is_transposon'].any():
                flags += 'T'
            # within 5 kb of end of contig
            if (int(row['start_position']) < length_from_end) or \
               (int(row['end_position']) > (scaffold_length_dict[row['scaffold']] - length_from_end)):
                flags += 'F'
            # if is_j:
            #     flags += 'J'
            flag_dict[gene] = flags
        # get 3 metabolic genes in a row flag
        for i in range(len(scaffold_annotations)):  # this needs to be fixed. Will only give B to middle of 3 genes.
            if 0 < i < (len(scaffold_annotations) - 1):
                gene = scaffold_annotations.index[i]
                gene_flags = flag_dict[gene]
                previous_gene = scaffold_annotations.index[i - 1]
                previous_gene_flags = flag_dict[previous_gene]
                next_gene = scaffold_annotations.index[i + 1]
                next_gene_flags = flag_dict[next_gene]
                if 'M' in previous_gene_flags and 'M' in gene_flags and 'M' in next_gene_flags:
                    if 'B' not in flag_dict[previous_gene]:
                        flag_dict[previous_gene] += 'B'
                    if 'B' not in flag_dict[gene]:
                        flag_dict[gene] += 'B'
                    if 'B' not in flag_dict[next_gene]:
                        flag_dict[next_gene] += 'B'
    return flag_dict


def get_amg_ids(amg_frame):
    ko_amgs = {j.strip() for i in amg_frame['KO'].dropna() for j in i.strip().split(';')}
    ec_amgs = {j.strip() for i in amg_frame['EC'].dropna() for j in i.strip().split(';')}
    pfam_amgs = {j.strip() for i in amg_frame['PFAM'].dropna() for j in i.strip().split(';')}
    # Put this back in once python 3.10 is more popular
    #return ko_amgs | ec_amgs | pfam_amgs
    return ko_amgs.union(ec_amgs, pfam_amgs)


def get_virsorter2_affi_contigs_name(scaffold):
    virsorter_scaffold_name = scaffold.replace('|', '_')
    return virsorter_scaffold_name


def get_virsorter_original_affi_contigs_name(scaffold):
    prophage_match = re.search(r'_gene_\d*_gene_\d*-\d*-\d*-cat_[123456]$', scaffold)
    plain_match = re.search(r'-cat_[123456]$', scaffold)
    if prophage_match is not None:
        virsorter_scaffold_name = scaffold[:prophage_match.start()]
    elif plain_match is not None:
        virsorter_scaffold_name = scaffold[:plain_match.start()]
    else:
        raise ValueError("Format is not VIRSorter2, and can't find VIRSorter "
                         "endings on fasta header: %s" % scaffold)
    return virsorter_scaffold_name


def get_virsorter_affi_contigs_name(scaffold):
    """
    :param scaffold: A raw string from the fasta header
    :returns: A string returned from the appropriate parser
    """
    virsorter2_match = re.search(r'\|\|', scaffold)
    if virsorter2_match is not None:
        return get_virsorter2_affi_contigs_name(scaffold)
    return get_virsorter_original_affi_contigs_name(scaffold)


def add_dramv_scores_and_flags(annotations, db_handler, logger, virsorter_hits=None, input_fasta=None):
    # setting up scoring viral genes
    amg_database_frame = pd.read_csv(db_handler.config['dram_sheets']['amg_database'], sep='\t')
    genome_summary_form = pd.read_csv(db_handler.config['dram_sheets']['genome_summary_form'], sep='\t', index_col=0)
    genome_summary_form = genome_summary_form.loc[genome_summary_form.potential_amg]

    # add auxiliary score
    if virsorter_hits is not None:
        gene_virsorter_category_dict = dict()
        gene_auxiliary_score_dict = dict()
        for scaffold, dram_frame in annotations.groupby('scaffold'):
            virsorter_scaffold_name = get_virsorter_affi_contigs_name(scaffold)
            virsorter_frame = virsorter_hits.loc[virsorter_hits.name == virsorter_scaffold_name]
            gene_order = get_gene_order(dram_frame, virsorter_frame)
            gene_virsorter_category_dict.update({dram_gene: virsorter_category for dram_gene, _, virsorter_category in
                                                 gene_order if dram_gene is not None})
            gene_auxiliary_score_dict.update(calculate_auxiliary_scores(gene_order))
        annotations['virsorter_category'] = pd.Series(gene_virsorter_category_dict)
        annotations['auxiliary_score'] = pd.Series(gene_auxiliary_score_dict)

    # get metabolic flags
    scaffold_length_dict = {seq.metadata['id']: len(seq) for seq in read_sequence(input_fasta, format='fasta')}
    metabolic_genes = set(genome_summary_form.index)
    if 'pfam_hits' in annotations:
        annotations['is_transposon'] = [is_transposon(i) for i in annotations['pfam_hits']]
    else:
        annotations['is_transposon'] = False

    amgs = get_amg_ids(amg_database_frame)
    verified_amgs = get_amg_ids(amg_database_frame.loc[amg_database_frame.verified])
    annotations['amg_flags'] = pd.Series(get_metabolic_flags(annotations, metabolic_genes, amgs,
                                                             verified_amgs, scaffold_length_dict,
                                                             logger))

    # downgrade B flag auxiliary scores
    if virsorter_hits is not None:
        annotations['auxiliary_score'] = pd.Series({gene: (4 if 'B' in row['amg_flags'] and row['auxiliary_score'] < 4
                                                           else row['auxiliary_score'])
                                                    for gene, row in annotations.iterrows()})

    return annotations

def annotate_vgfs(input_fasta, virsorter_affi_contigs=None, output_dir='.', min_contig_size=2500, split_contigs=False,
                  prodigal_mode='meta', trans_table='11', bit_score_threshold=60, rbh_bit_score_threshold=350,
                  custom_db_name=(), custom_fasta_loc=(), custom_hmm_loc=(), custom_hmm_cutoffs_loc=(),
                  custom_hmm_name=(), use_uniref=False, kofam_use_dbcan2_thresholds=False, skip_trnascan=False,
                  keep_tmp_dir=True, low_mem_mode=False, threads=10, verbose=True, config_loc:str=None ):
    mkdir(output_dir)
    log_file_path = path.join(output_dir, "annotate.log")
    logger = logging.getLogger('annotation_log')
    setup_logger(logger, log_file_path)

    # check inputs
    prodigal_modes = ['train', 'meta', 'single']
    if prodigal_mode not in prodigal_modes:
        raise ValueError('Prodigal mode must be one of %s.' % ', '.join(prodigal_modes))
    elif prodigal_mode in ['normal', 'single']:
        warnings.warn('When running prodigal in single mode your bins must have long contigs (average length >3 Kbp), '
                      'be long enough (total length > 500 Kbp) and have very low contamination in order for prodigal '
                      'training to work well.')

    # get database locations
    db_handler = DatabaseHandler(logger, config_loc=config_loc)
    db_handler.filter_db_locs(low_mem_mode=low_mem_mode,
                              use_uniref=use_uniref,
                              use_vogdb=True,
                              master_list=VMAG_DBS_TO_ANNOTATE)
    if virsorter_affi_contigs is not None:
        virsorter_hits = get_virsorter_hits(virsorter_affi_contigs)
    else:
        virsorter_hits = None

    # Check fastas
    perform_fasta_checks([input_fasta], logger)

    logger.info(f"Starting the Viral Annotation of Genes with database configuration: \n {db_handler.get_settings_str()}")

    if split_contigs:
        # split sequences into separate fastas
        contig_dir = path.join(output_dir, 'vMAGs')
        mkdir(contig_dir)

        contig_locs = list()
        for seq in read_sequence(input_fasta, format='fasta'):
            if len(seq) >= min_contig_size:
                if '=' in seq.metadata['id'] or ';' in seq.metadata['id']:
                    raise ValueError('FASTA headers must not have = or ; before the first space (%s). To run DRAM-v '
                                     'you must rerun VIRSorter with = and ; removed from the headers or run DRAM-v.py '
                                     'remove_bad_characters and then rerun DRAM-v' % seq.metadata['id'])
                if virsorter_hits is not None:
                    if get_virsorter_affi_contigs_name(seq.metadata['id']) not in virsorter_hits['name'].values:
                        raise ValueError("No virsorter calls found in %s for scaffold %s from input fasta" %
                                         (virsorter_affi_contigs, seq.metadata['id']))
                contig_loc = path.join(contig_dir, '%s.fasta' % seq.metadata['id'])
                write_sequence((i for i in [seq]), format='fasta', into=contig_loc)
                contig_locs.append(contig_loc)
    else:
        contig_locs = [input_fasta]

    # annotate vMAGs
    rename_bins = False
    annotations = annotate_fastas(
        fasta_locs=contig_locs,
        output_dir=output_dir, 
        db_handler=db_handler,
        logger=logger,
        min_contig_size=min_contig_size,
        prodigal_mode=prodigal_mode,
        trans_table=trans_table,
        bit_score_threshold=bit_score_threshold,
        rbh_bit_score_threshold=rbh_bit_score_threshold,
        custom_db_name=custom_db_name,
        custom_fasta_loc=custom_fasta_loc,
        custom_hmm_name=custom_hmm_name,
        custom_hmm_loc=custom_hmm_loc,
        custom_hmm_cutoffs_loc=custom_hmm_cutoffs_loc,
        kofam_use_dbcan2_thresholds=kofam_use_dbcan2_thresholds,
        skip_trnascan=skip_trnascan,
        rename_bins=rename_bins,
        keep_tmp_dir=keep_tmp_dir,
        threads=threads,
        verbose=verbose)
    logging.info('Annotations complete, assigning auxiliary scores and flags')

    annotations = add_dramv_scores_and_flags(annotations, db_handler, logger, virsorter_hits, input_fasta)

    # write annotations
    annotations.to_csv(path.join(output_dir, 'annotations.tsv'), sep='\t')

    logging.info("Completed annotations")




"""
Contains most of the backend for the DRAM_setup.py script, used to setup databases for each user.
"""
from os import path, mkdir
from datetime import datetime
from shutil import move, rmtree, copyfile
from glob import glob
import gzip
from collections import defaultdict
import logging
import tarfile

from skbio import read as read_sequence
from skbio import write as write_sequence

from mag_annotator.database_handler import DatabaseHandler
from mag_annotator.utils import run_process, make_mmseqs_db, download_file, merge_files, remove_prefix, remove_suffix, setup_logger

NUMBER_OF_VIRAL_FILES = 2
DEFAULT_DBCAN_RELEASE = '11'
DEFAULT_DBCAN_DATE = '08062022'
DEFAULT_UNIREF_VERSION = '90'
DEFAULT_VOGDB_VERSION = 'latest'
DFLT_OUTPUT_DIR = '.'
LOGGER = logging.getLogger("database_processing.log")
DEFAULT_MMMSPRO_DB_NAME = 'db'


KOFAM_CITATION = ("T. Aramaki, R. Blanc-Mathieu, H. Endo, K. Ohkubo, M. Kanehisa"
                  ", S. Goto, and H. Ogata, \"Kofamkoala: Kegg ortholog assignme"
                  "nt based on profile hmm and adaptive score threshold,\" Bioin"
                  "formatics, vol. 36, no. 7, pp. 2251–2252, 2020."
                  )
VIRAL_REFSEQ_CITATION = ("J. R. Brister, D. Ako-Adjei, Y. Bao, and O. Blinkova, "
                         "\"Ncbi viral genomes resource,\" Nucleic acids researc"
                         "h, vol. 43, no. D1, pp. D571–D577, 2015. [3] M. Kanehi"
                         "sa, M. Furumichi, Y. Sato, M. Ishiguro-Watanabe, and M"
                         ". Tan-abe, \"Kegg: integrating viruses and cellular or"
                         "ganisms,\" Nucleic acids research, vol. 49, no. D1, pp"
                         ". D545–D551, 2021."
                 )
KEGG_CITATION = (" M. Kanehisa, M. Furumichi, Y. Sato, M. Ishiguro-Watanabe, and"
                 " M. Tanabe, \"Kegg: integrating viruses and cellular organisms"
                 ",\" Nucleic acids research, vol. 49, no. D1, pp. D545–D551, 20"
                 "21."
                 )
PFAM_CITATION = ("J. Mistry, S. Chuguransky, L. Williams, M. Qureshi, G. A. Sal"
                 "azar, E. L. Sonnhammer, S. C. Tosatto, L. Paladin, S. Raj, L."
                 " J. Richardson et al., \"Pfam: The protein families database "
                 "in 2021,\" Nucleic acids research, vol. 49, no. D1, pp. D412–"
                 "D419, 2021."
                 )
PEPTIDASE_CITATION = ("N. D. Rawlings, A. J. Barrett, P. D. Thomas, X. Huang, A"
                      ". Bateman, and R. D. Finn, \"The merops database of prot"
                      "eolytic enzymes, their substrates and inhibitors in 2017"
                      " and a comparison with peptidases in the panther databas"
                      "e,\" Nucleic acids research, vol. 46, no. D1, pp. D624–D"
                      "632, 2018."
                      )
VOGDB_CITATION = ("J. Thannesberger, H.-J. Hellinger, I. Klymiuk, M.-T. Kastner"
                  ", F. J. Rieder, M. Schneider, S. Fister, T. Lion, K. Kosulin"
                  ", J. Laengle et al., \"Viruses comprise an extensive pool of"
                  " mobile genetic elements in eukaryote cell cultures and huma"
                  "n clinical samples,\" The FASEB Journal, vol. 31, no. 5, pp."
                  " 1987–2000, 2017."
                  )
UNIREF_CITATION = ("Y. Wang, Q. Wang, H. Huang, W. Huang, Y. Chen, P. B. McGarv"
                  "ey, C. H. Wu, C. N. Arighi, and U. Consortium, \"A crowdsour"
                  "cing open platform for literature curation in uniprot,\" PLo"
                  "S Biology, vol. 19, no. 12, p. e3001464, 2021."
                   )
DBCAN_CITATION = ("Y. Yin, X. Mao, J. Yang, X. Chen, F. Mao, and Y. Xu, \"dbcan"
                  ": a web resource for automated carbohydrate-active enzyme an"
                  "notation,\" Nucleic acids research, vol. 40, no. W1, pp. W44"
                  "5–W451, 2012."
                  )
DRAM_CITATION = ("M. Shaffer, M. A. Borton, B. B. McGivern, A. A. Zayed, S. L. "
                 "La Rosa, L. M. Solden, P. Liu, A. B. Narrowe, J. Rodríguez-Ra"
                 "mos, B. Bolduc et al., \"Dram for distilling microbial metabo"
                 "lism to automate the curation of microbiome function,\" Nucle"
                 "ic acids research, vol. 48, no. 16, pp. 8883–8900, 2020."
                 )
# check 
# if dbcapfam is down, raise appropriate error
# TODO: upgrade to pigz?


def get_iso_date():
    return datetime.today().strftime('%Y%m%d')


def download_pfam_hmm(output_dir='.', logger=LOGGER, verbose=True):
    pfam_hmm = path.join(output_dir, 'Pfam-A.hmm.dat.gz')
    url = 'ftp://ftp.ebi.ac.uk/pub/databases/Pfam/current_release/Pfam-A.hmm.dat.gz'
    url_http = 'http://ftp.ebi.ac.uk/pub/databases/Pfam/current_release/Pfam-A.hmm.dat.gz'
    logger.debug(f"Downloading Pfam from: {url}")
    download_file(url, pfam_hmm, logger, alt_urls=[url_http], verbose=verbose)
    return pfam_hmm


def download_dbcan(output_dir='.', logger=LOGGER, dbcan_hmm=None, version=DEFAULT_DBCAN_RELEASE, verbose=True):
    dbcan_hmm = path.join(output_dir, f"dbCAN-HMMdb-V{version}.txt" )
    if int(version) < int(DEFAULT_DBCAN_RELEASE):
        link_path = f"http://bcb.unl.edu/dbCAN2/download/Databases/dbCAN-HMMdb-V{version}.txt"
    else:
        link_path = f"http://bcb.unl.edu/dbCAN2/download/dbCAN-HMMdb-V{version}.txt"

    logger.debug(f"Downloading dbCAN from: {link_path}")
    download_file(link_path, dbcan_hmm, logger, verbose=verbose)
    return dbcan_hmm


def download_dbcan_fam_activities (output_dir='.', logger=LOGGER, version=DEFAULT_DBCAN_RELEASE, upload_date=DEFAULT_DBCAN_DATE, 
                                verbose=True):
    dbcan_fam_activities = path.join(output_dir, f'CAZyDB.{upload_date}.fam-activities.txt')
    url = f"https://bcb.unl.edu/dbCAN2/download/Databases/V{version}/CAZyDB.{upload_date}.fam-activities.txt"
    logger.info(f"Downloading dbCAN family activities from : {url}")
    download_file(url, dbcan_fam_activities, logger, verbose=verbose)
    return dbcan_fam_activities


def download_dbcan_subfam_ec(output_dir='.', logger=LOGGER, version=DEFAULT_DBCAN_RELEASE, upload_date=DEFAULT_DBCAN_DATE, verbose=True):
    dbcan_subfam_ec = path.join(output_dir, f"CAZyDB.{upload_date}.fam.subfam.ec.txt")
    url = (f"https://bcb.unl.edu/dbCAN2/download/Databases/"
                 f"V{version}/CAZyDB.{upload_date}.fam.subfam.ec.txt")
    logger.info(f"Downloading dbCAN sub-family encumber from : {url}")
    download_file(url, dbcan_subfam_ec, logger, verbose=verbose)
    return dbcan_subfam_ec


def download_kofam_hmm(output_dir='.', logger=LOGGER, verbose=False):
    kofam_profile_tar_gz = path.join(output_dir, 'kofam_profiles.tar.gz')
    url = 'ftp://ftp.genome.jp/pub/db/kofam/profiles.tar.gz'  
    url_http = 'https://www.genome.jp/ftp/db/kofam/profiles.tar.gz'
    download_file(url, kofam_profile_tar_gz, logger, alt_urls=[url_http], verbose=verbose)
    return kofam_profile_tar_gz

def generate_modified_kegg_fasta(kegg_fasta, gene_ko_link_loc=None):
    """
    Takes kegg fasta file and gene ko link file, adds kos not already in headers to headers
    Whish I knew about this, oh well I may split this out.
    """
    genes_ko_dict = defaultdict(list)
    if gene_ko_link_loc is not None:
        if gene_ko_link_loc.endswith('.gz'):
            gene_ko_link_fh = gzip.open(gene_ko_link_loc, 'rt')
        else:
            gene_ko_link_fh = open(gene_ko_link_loc)
        for line in gene_ko_link_fh:
            gene, ko = line.strip().split()
            genes_ko_dict[gene].append(remove_prefix(ko, 'ko:'))
    for seq in read_sequence(kegg_fasta, format='fasta'):
        new_description = seq.metadata['description']
        for ko in genes_ko_dict[seq.metadata['id']]:
            if ko not in new_description:
                new_description += '; %s' % ko
        seq.metadata['description'] = new_description
        yield seq



def process_kegg(kegg_loc, output_dir, logger, gene_ko_link_loc=None, download_date=None, 
                    threads=10, verbose=True):
    if download_date is None:
        download_date = get_iso_date()
    if gene_ko_link_loc is not None:
        # add KOs to end of header where KO is not already there
        kegg_mod_loc = path.join(output_dir, 'kegg.mod.fa')
        write_sequence(generate_modified_kegg_fasta(kegg_loc, gene_ko_link_loc),
                       format='fasta', into=kegg_mod_loc)
    else:
        kegg_mod_loc = kegg_loc
    # make mmseqsdb from modified kegg fasta
    kegg_mmseqs_db = path.join(output_dir, 'kegg.%s.mmsdb' % download_date)
    make_mmseqs_db(kegg_mod_loc, kegg_mmseqs_db, logger, create_index=True, threads=threads, verbose=verbose)
    LOGGER.info('KEGG database processed')
    return {'kegg': kegg_mmseqs_db}


def process_kofam_hmm(kofam_profile_tar_gz, output_dir=DFLT_OUTPUT_DIR, logger=LOGGER, threads=1, verbose=False):
    kofam_profiles = path.join(output_dir, 'kofam_profiles')
    mkdir(kofam_profiles)
    run_process(['tar', '-xzf', kofam_profile_tar_gz, '-C', kofam_profiles], logger, verbose=verbose)
    merged_kofam_profiles = path.join(output_dir, 'kofam_profiles.hmm')
    merge_files(glob(path.join(kofam_profiles, 'profiles', '*.hmm')), merged_kofam_profiles)
    run_process(['hmmpress', '-f', merged_kofam_profiles], logger, verbose=verbose)
    LOGGER.info('KOfam database processed')
    return {'kofam_hmm': merged_kofam_profiles}


def download_kofam_ko_list(output_dir='.', logger=LOGGER, verbose=False):
    kofam_ko_list_gz = path.join(output_dir, 'kofam_ko_list.tsv.gz')
    url = 'ftp://ftp.genome.jp/pub/db/kofam/ko_list.gz'
    url_http = 'https://www.genome.jp/ftp/db/kofam/ko_list.gz'
    download_file(url, kofam_ko_list_gz, logger, alt_urls=[url_http], verbose=verbose)
    return kofam_ko_list_gz


def download_pfam(output_dir='.', logger=LOGGER, verbose=True):
    pfam_full_zipped = path.join(output_dir, 'Pfam-A.full.gz')
    url = 'ftp://ftp.ebi.ac.uk/pub/databases/Pfam/current_release/Pfam-A.full.gz'
    http_url = 'http://ftp.ebi.ac.uk/pub/databases/Pfam/current_release/Pfam-A.full.gz'
    download_file(url, pfam_full_zipped, logger, alt_urls=[http_url], verbose=verbose)
    return pfam_full_zipped


def download_viral(output_dir='.', logger=LOGGER, viral_files=NUMBER_OF_VIRAL_FILES, verbose=True):
    """Can only download newest version"""
    # download all of the viral protein files, need to know the number of files
    # TODO: Make it so that you don't need to know number of viral files in refseq viral
    faa_base_name = 'viral.%s.protein.faa.gz'
    viral_faa_glob = path.join(output_dir, faa_base_name % '*')
    for number in range(viral_files):
        number += 1
        url = 'ftp://ftp.ncbi.nlm.nih.gov/refseq/release/viral/viral.%s.protein.faa.gz' % number
        url_http = 'https://ftp.ncbi.nlm.nih.gov/refseq/release/viral/viral.%s.protein.faa.gz' % number
        output_name= path.join(output_dir, faa_base_name % number)
        download_file(url, output_name, logger, alt_urls=[url_http], verbose=verbose)
    # then merge files from above
    merged_viral_faas = path.join(output_dir, 'viral.merged.protein.faa.gz')
    run_process(['cat %s > %s' % (' '.join(glob(viral_faa_glob)), merged_viral_faas)], logger, shell=True)
    return merged_viral_faas


def download_uniref(output_dir='.', logger=LOGGER, version=DEFAULT_UNIREF_VERSION, 
                    threads=10, verbose=True):
    uniref_fasta_zipped = path.join(output_dir, 'uniref%s.fasta.gz' % version)
    url = 'ftp://ftp.uniprot.org/pub/databases/uniprot/uniref/uniref%s/uniref%s.fasta.gz' % \
          (version, version)
    url_http = 'https://ftp.uniprot.org/pub/databases/uniprot/uniref/uniref%s/uniref%s.fasta.gz' % \
                 (version, version)
    download_file(url, uniref_fasta_zipped, logger, alt_urls=[url_http], verbose=verbose)
    return uniref_fasta_zipped


def download_peptidase(output_dir='.', logger=LOGGER, verbose=True):
    save_name = path.join(output_dir, 'merops_peptidases_nr.faa')
    url = 'ftp://ftp.ebi.ac.uk/pub/databases/merops/current_release/pepunit.lib'
    url_http = 'https://ftp.ebi.ac.uk/pub/databases/merops/current_release/pepunit.lib'
    download_file(url, save_name, logger, alt_urls=[url_http], verbose=verbose)
    return save_name

def download_vogdb(output_dir='.', logger=LOGGER, version=DEFAULT_VOGDB_VERSION, verbose=True):
    vog_hmm_targz = path.join(output_dir, 'vog.hmm.tar.gz')
    vogdb_url = f'http://fileshare.csb.univie.ac.at/vog/{version}/vog.hmm.tar.gz'
    download_file(vogdb_url, vog_hmm_targz, logger, verbose=verbose)
    return vog_hmm_targz


def process_kofam_ko_list(kofam_ko_list_gz, output_dir='.', logger=LOGGER, threads=1, verbose=False):
    # TODO: fix this so that it is gunzipped to the path
    kofam_ko_list = path.join(output_dir, 'kofam_ko_list.tsv')
    run_process(['gunzip', '-c', kofam_ko_list_gz], logger, save_output=kofam_ko_list, verbose=verbose)
    LOGGER.info('KOfam ko list processed')
    return {'kofam_ko_list': kofam_ko_list}


def process_uniref(uniref_fasta_zipped, output_dir='.', logger=LOGGER, 
                   version=DEFAULT_UNIREF_VERSION, threads=10,
                   verbose=True):
    uniref_mmseqs_db = path.join(output_dir, 'uniref%s.%s.mmsdb' % (version, get_iso_date()))
    make_mmseqs_db(uniref_fasta_zipped, uniref_mmseqs_db, logger, create_index=True, threads=threads, verbose=verbose)
    LOGGER.info('UniRef database processed')
    return {'uniref': uniref_mmseqs_db}


def process_mmspro(full_alignment, output_dir, logger=LOGGER, db_name=DEFAULT_MMMSPRO_DB_NAME, threads=10, verbose=True):
    mmseqs_msa = path.join(output_dir, '%s.mmsmsa' % db_name)
    run_process(['mmseqs', 'convertmsa', full_alignment, mmseqs_msa], logger, verbose=verbose)
    mmseqs_profile = path.join(output_dir, '%s.mmspro' % db_name)
    run_process(['mmseqs', 'msa2profile', mmseqs_msa, mmseqs_profile, '--match-mode', '1', '--threads', str(threads)], 
                logger,
                verbose=verbose)
    tmp_dir = path.join(output_dir, 'tmp')
    run_process(['mmseqs', 'createindex', mmseqs_profile, tmp_dir, '-k', '5', '-s', '7', '--threads', str(threads)], 
                logger,
                verbose=verbose)
    return mmseqs_profile


def process_pfam(pfam_full_zipped, output_dir='.', logger=LOGGER, threads=10, verbose=True):
    pfam_profile = process_mmspro(pfam_full_zipped, output_dir, logger, 'pfam', threads, verbose)
    LOGGER.info('PFAM database processed')
    return {'pfam': pfam_profile}


def process_dbcan(input, output_dir, logger=LOGGER, verbose=True, threads=1):
    output = path.join(output_dir, path.basename(input))
    move(input, output)
    run_process(['hmmpress', '-f', output], logger, verbose=verbose)
    LOGGER.info('dbCAN database processed')
    return {'dbcan': output}


def process_viral(merged_viral_faas, output_dir='.', logger=LOGGER, viral_files=NUMBER_OF_VIRAL_FILES, threads=10, verbose=True):
    refseq_viral_mmseqs_db = path.join(output_dir, 'refseq_viral.%s.mmsdb' % get_iso_date())
    make_mmseqs_db(merged_viral_faas, refseq_viral_mmseqs_db, logger, create_index=True, threads=threads, verbose=verbose)
    LOGGER.info('RefSeq viral database processed')
    return {'viral': refseq_viral_mmseqs_db}



def process_peptidase(peptidase_faa, output_dir='.', logger=LOGGER, threads=10, verbose=True):
    peptidase_mmseqs_db = path.join(output_dir, 'peptidases.%s.mmsdb' % get_iso_date())
    make_mmseqs_db(peptidase_faa, peptidase_mmseqs_db, logger, create_index=True, threads=threads, verbose=verbose)
    LOGGER.info('MEROPS database processed')
    return {'peptidase': peptidase_mmseqs_db}


def process_vogdb(vog_hmm_targz, output_dir='.', logger=LOGGER, version=DEFAULT_VOGDB_VERSION, threads=1, verbose=True):
    hmm_dir = path.join(output_dir, 'vogdb_hmms')
    mkdir(hmm_dir)
    vogdb_targz = tarfile.open(vog_hmm_targz)
    vogdb_targz.extractall(hmm_dir)
    vog_hmms = path.join(output_dir, f'vog_{version}_hmms.txt')
    merge_files(glob(path.join(hmm_dir, 'VOG*.hmm')), vog_hmms)
    run_process(['hmmpress', '-f', vog_hmms], logger, verbose=verbose) 
    LOGGER.info('VOGdb database processed')
    return {'vogdb': vog_hmms}


def download_vog_annotations(output_dir, logger=LOGGER, version=DEFAULT_VOGDB_VERSION, verbose=True):
    vog_annotations = path.join(output_dir, 'vog_annotations_%s.tsv.gz' % version)
    download_file('http://fileshare.csb.univie.ac.at/vog/%s/vog.annotations.tsv.gz' % version,
                  vog_annotations, logger, verbose=verbose)
    return vog_annotations


def download_genome_summary_form(output_dir, logger=LOGGER, branch='master', verbose=True):
    genome_summary_form = path.join(output_dir, 'genome_summary_form.%s.tsv' % get_iso_date())
    download_file('https://raw.githubusercontent.com/WrightonLabCSU/DRAM/%s/data/genome_summary_form.tsv' % branch,
                  genome_summary_form, logger, verbose=verbose)
    return genome_summary_form


def download_module_step_form(output_dir, logger=LOGGER, branch='master', verbose=True):
    function_heatmap_form = path.join(output_dir, 'module_step_form.%s.tsv' % get_iso_date())
    download_file('https://raw.githubusercontent.com/WrightonLabCSU/DRAM/%s/data/module_step_form.tsv' % branch,
                  function_heatmap_form, logger, verbose=verbose)
    return function_heatmap_form


def download_etc_module_database(output_dir, logger=LOGGER, branch='master', verbose=True):
    etc_module_database = path.join(output_dir, 'etc_mdoule_database.%s.tsv' % get_iso_date())
    download_file('https://raw.githubusercontent.com/WrightonLabCSU/DRAM/%s/data/etc_module_database.tsv' % branch,
                  etc_module_database, logger, verbose=verbose)
    return etc_module_database


def download_function_heatmap_form(output_dir, logger=LOGGER, branch='master', verbose=True):
    function_heatmap_form = path.join(output_dir, 'function_heatmap_form.%s.tsv' % get_iso_date())
    download_file('https://raw.githubusercontent.com/WrightonLabCSU/DRAM/%s/data/function_heatmap_form.tsv' % branch,
                  function_heatmap_form, logger, verbose=verbose)
    return function_heatmap_form


def download_amg_database(output_dir, logger=LOGGER, branch='master', verbose=True):
    amg_database = path.join(output_dir, 'amg_database.%s.tsv' % get_iso_date())
    download_file('https://raw.githubusercontent.com/WrightonLabCSU/DRAM/%s/data/amg_database.tsv' % branch, 
                  amg_database, logger, verbose=verbose,)
    return amg_database


def check_file_exists(*paths):
    for i in paths:
        if i is None:
            continue
        elif path.isfile(i):
            continue
        else:
            raise ValueError(f"Database location does not exist: {i}")


def prepare_databases(output_dir, loggpath=None, kegg_loc=None, gene_ko_link_loc=None, kofam_hmm_loc=None,
                      kofam_ko_list_loc=None, kegg_download_date=None, uniref_loc=None,
                      uniref_version=DEFAULT_UNIREF_VERSION, pfam_loc=None, pfam_hmm_loc=None,
                      dbcan_loc=None, dbcan_fam_activities:str=None, dbcan_subfam_ec:str=None, dbcan_version=DEFAULT_DBCAN_RELEASE,
                      dbcan_date=DEFAULT_DBCAN_DATE, viral_loc=None, peptidase_loc=None,
                      vogdb_loc=None, vogdb_version=DEFAULT_VOGDB_VERSION, vog_annotations=None,
                      genome_summary_form_loc=None, module_step_form_loc=None,
                      etc_module_database_loc=None, function_heatmap_form_loc=None,
                      amg_database_loc=None,
                      number_of_viral_files=NUMBER_OF_VIRAL_FILES,
                      skip_uniref=False, keep_database_files=False,
                      branch='master', threads=10, verbose=True, select_db=None, clear_config=False):


    dram_settings = {
        'kegg':                  {'name': 'KEGG db', 
                                  'description_db_updated': 'Unknown, or Never', 
                                  'citation': KEGG_CITATION},
        'gene_ko_link':          {'name': 'KEGG Gene KO link', 'citation': KEGG_CITATION},
        "kofam_hmm":             {'name': 'KOfam db', 'citation': KOFAM_CITATION},
        "kofam_ko_list":         {'name': 'KOfam KO list', 'citation': KOFAM_CITATION},
        'uniref':                {'name': 'UniRef db', 
                                  'description_db_updated': 'Unknown, or Never', 
                                  'citation': UNIREF_CITATION},
        'pfam':                  {'name': 'Pfam db', 'citation': PFAM_CITATION},
        "pfam_hmm":          {'name': 'Pfam hmm dat', 
                                  'description_db_updated': 'Unknown, or Never', 
                                  'citation': PFAM_CITATION},
        "dbcan":                 {'name': 'dbCAN db', 'citation': DBCAN_CITATION},
        "dbcan_fam_activities":  {'name': 'dbCAN family activities', 'citation': DBCAN_CITATION}, 
        "dbcan_subfam_ec":       {'name': 'dbCAN subfamily EC numbers', 'citation': DBCAN_CITATION},
        "vogdb":                 {'name': 'VOGDB db', 'citation': VOGDB_CITATION},
        "vog_annotations":       {'name': 'VOG annotations', 
                                  'description_db_updated': 'Unknown, or Never', 
                                  'citation': VOGDB_CITATION},
        "viral":          {'name': 'RefSeq Viral db', 
                                  'description_db_updated': 'Unknown, or Never', 
                                  'citation': VIRAL_REFSEQ_CITATION},
        "peptidase":             {'name': 'MEROPS peptidase db', 
                                  'description_db_updated': 'Unknown, or Never', 
                                  'citation': PEPTIDASE_CITATION}, 
        "genome_summary_form":   {'name': 'Genome summary form'},
        "module_step_form":      {'name': 'Module step form'},
        "function_heatmap_form": {'name': 'Function heatmap form'},
        "amg_database":          {'name': 'AMG database'},
        "etc_module_database":   {'name': 'ETC module database'},
    }
    database_settings= {
        'kegg':                  {},
        "gene_ko_link":          {},
        "kofam_hmm":             {},
        "kofam_ko_list":         {},
        'uniref':                {'version': uniref_version},
        'pfam':                  {},
        "pfam_hmm":          {},
        "dbcan":                 {'version': dbcan_version},
        "dbcan_fam_activities":  {'version': dbcan_version, 'upload_date': dbcan_date}, 
        "dbcan_subfam_ec":       {'version': dbcan_version, 'upload_date': dbcan_date},
        "vogdb":                 {'version': vogdb_version},
        "vog_annotations":       {'version': vogdb_version},
        "viral":                 {},
        "peptidase":             {}, 
        "genome_summary_form":   {"branch": branch},
        "module_step_form":      {"branch": branch},
        "function_heatmap_form": {"branch": branch},
        "amg_database":          {"branch": branch},
        "etc_module_database":   {"branch": branch},
    }
    process_settings = {
        'kegg': {},
        "gene_ko_link": {},
        "kofam_hmm": {},
        "kofam_ko_list": {},
        'uniref': {"version": uniref_version},
        'pfam': {},
        "pfam_hmm": {},
        "dbcan": {},
        "dbcan_fam_activities": {}, 
        "dbcan_subfam_ec": {},
        "vogdb": {},
        "vog_annotations": {},
        "viral": {'viral_files': number_of_viral_files},
        "peptidase": {}, 
        "genome_summary_form": {},
        "module_step_form": {},
        "function_heatmap_form": {},
        "amg_database": {},
        "etc_module_database": {},
    }

    # setup temp, logging, and db_handler
    if not path.isdir(output_dir):
        mkdir(output_dir)
    temporary = path.join(output_dir, 'database_files')
    mkdir(temporary)
    main_log = path.join(output_dir, 'database_processing.log')
    setup_logger(LOGGER, *[(main_log, loggpath) if loggpath is not None else main_log])
    db_handler = DatabaseHandler(logger=LOGGER)
    if clear_config or select_db is None:
        db_handler.clear_config()

    db_handler.config['log_path'] = main_log
    db_handler.write_config()
    LOGGER.info('Starting the process of downloading data')

    if skip_uniref:
        LOGGER.info('Skipping UniRef')
        del database_settings['uniref']
    
    
    locs = {remove_suffix(i, '_loc'):j for i, j in locals().items() if i.endswith('_loc') and j is not None}
    download_functions = {i[9:]:j for i,j in globals().items() if callable(j) and i.startswith('download_')}
    process_functions = {i[8:]:j for i,j in globals().items() if callable(j) and i.startswith('process_')}
    functions = {i:j for i,j in globals().items() if callable(j) and i.startswith('download_')}

    # Check any specified paths exist
    missing_user_inputs = [f"{i} at {j}" for i, j in locs.items() if not path.exists(j)]
    if len(missing_user_inputs) > 1:
        raise ValueError(f"The fallowing user provided paths don't seem to exist: {', '.join(missing_user_inputs)}")

    un_obtainable = [i for i in database_settings if i not in locs and i not in download_functions]

    for i in un_obtainable:
        LOGGER.info(f"The {i}_loc argument was not used to specify a downloaded {i} file, and dram can not"
                     " download it its self. So it is assumed that the user wants to set up DRAM without it")
        del database_settings[i]


    # check inputs
    if skip_uniref and uniref_loc is not None:
        raise ValueError('Cannot skip UniRef processing and provide a location of UniRef.'
                         ' Skipping UniRef will cause provided UniRef file to not be used.')
    
    if select_db is not None:
        miss_name = [i for i in select_db if i not in database_settings]
        user_inputs = [i for i in locs if i not in select_db]
        if len(miss_name) > 0:
            LOGGER.error("Only the databases in the db list can be pased to select_db, "
                            f"you passed {miss_name} which is/are not in the list,"
                            f" select from {list(database_settings.keys())}.")
            raise ValueError('Bad user input, see log')
        
        if len(user_inputs) > 0:
            LOGGER.error(f"The user provided location for {user_inputs}, but required it not be used by proving"
                             f" the select_db argument for other databases. This would suggest that the"
                             " user may have made a mistake and so this error is rased.")
            raise ValueError('Bad user input, see log')

        database_settings = {i:database_settings[i] for i in select_db}


    LOGGER.info('Database preparation started')

    # Download DBs
    for i, j in database_settings.items():
        if locs.get(i) is None:
            LOGGER.info(f"Downloading {i}")
            if i in process_functions:
                locs[i] = download_functions[i](
                    temporary, LOGGER, **j, verbose=verbose)
            else:
                locs[i] = download_functions[i](
                    output_dir, LOGGER, **j, verbose=verbose)
            j['Download time'] = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
            j['Origin'] = "Downloaded by DRAM"
        else:
            j = {k: "Unknown" for k in j}
            j['Download time'] = "Unknown"
            j['Origin'] = "Provided by user"
            j['Original path'] = locs[i]
            if i not in process_functions:
                LOGGER.info(f"Copying {locs[i]} to output_dir")
                locs[i] = copyfile(locs[i], path.join(output_dir, path.basename(locs[i])))

    LOGGER.info("All raw data files were downloaded successfully")

    # Process databases
    for i in locs: 
        processed_locs = {} 
        if i in process_functions:
            LOGGER.info(f"Processing {i}")
            processed_locs = process_functions[i](locs[i], output_dir, LOGGER, 
                                                       threads=threads, verbose=verbose, **process_settings[i])
        else:
            processed_locs = {i:locs[i]}
        for k, v in processed_locs.items():
            final_dest = path.join(output_dir, path.basename(v))
            if v != final_dest:
                for db_file in glob('%s*' % v):
                    move(db_file, path.join(output_dir, path.basename(db_file)))
                v = path.join(output_dir, path.basename(v))
            # update_dram_forms the settings per OUTPUT fill, including the process_settings
            #  and database_settings, which are per input file.
            if db_handler.config.get('setup_info') is None:
                db_handler.config['setup_info'] = {}
            db_handler.config['setup_info'][k] = {**dram_settings[k], **process_settings[i], 
                                                  **database_settings[i]}
            db_handler.set_database_paths(**{f"{k}_loc":v})
            db_handler.write_config()
            LOGGER.info(f'Moved {k} to final destination, configuration updated')

    LOGGER.info('Populating the description db, this may take some time')
    db_handler.config['description_db'] = path.realpath(path.join(output_dir, 'description_db.sqlite'))
    db_handler.write_config()
    db_handler.populate_description_db(db_handler.config['description_db'], select_db, update_config=False)
    # todo make db handler such that the destruction on success write_config
    db_handler.write_config()
    LOGGER.info('DRAM description database populated')

    if not keep_database_files:
        rmtree(temporary)
    LOGGER.info('Database preparation completed')


def update_dram_forms(output_dir, branch='master'):
    if not path.isdir(output_dir):
        mkdir(output_dir)
    form_locs = dict()
    form_locs['genome_summary_form_loc'] = download_genome_summary_form(output_dir, branch)
    form_locs['module_step_form_loc'] = download_module_step_form(output_dir, branch)
    form_locs['etc_module_database_loc'] = download_etc_module_database(output_dir, branch)
    form_locs['function_heatmap_form_loc'] = download_function_heatmap_form(output_dir, branch)
    form_locs['amg_database_loc'] = download_amg_database(output_dir, branch)
    main_log = path.join(output_dir, 'database_processing.log')
    setup_logger(LOGGER, main_log)
    db_handler = DatabaseHandler(LOGGER)
    db_handler.set_database_paths(**form_locs)



import pytest
import os
import logging

from mag_annotator.utils import setup_logger
from mag_annotator.database_processing import generate_modified_kegg_fasta, process_kegg, process_mmspro

@pytest.fixture()
def phix_proteins():
    return os.path.join('tests', 'data', 'NC_001422.faa')


@pytest.fixture()
def fake_gene_link_loc():
    return os.path.join('tests', 'data', 'fake_gene_ko.link')

@pytest.fixture()
def logger(tmpdir):
    logger = logging.getLogger('test_log')
    setup_logger(logger)
    return logger


def test_generate_modified_kegg_fasta(phix_proteins, fake_gene_link_loc):
    seqs = generate_modified_kegg_fasta(phix_proteins,
                                        fake_gene_link_loc)
    seqs_dict = {seq.metadata['id']: seq for seq in seqs}
    assert 'K14021' in seqs_dict['NP_040705.1'].metadata['description']
    assert 'K04464' in seqs_dict['NP_040708.1'].metadata['description']


def test_process_kegg(tmpdir, phix_proteins, fake_gene_link_loc, logger):
    processed_kegg_dir = tmpdir.mkdir('process_kegg_test')
    kegg = process_kegg(phix_proteins, processed_kegg_dir, logger, fake_gene_link_loc, download_date='today')
    assert os.path.isfile(kegg['kegg'])
    assert os.path.isfile('%s_h' % kegg['kegg'])


def test_process_mmspro(tmpdir, logger):
    processed_mmspro = tmpdir.mkdir('process_mmspro')
    process_mmspro(os.path.join('tests', 'data', 'Pfam-A_subset.full.gz'), processed_mmspro, logger, 'fake', 1, False)
    assert os.path.isfile(os.path.join(processed_mmspro, 'fake.mmspro'))
    assert os.path.isfile(os.path.join(processed_mmspro, 'fake.mmspro_h'))
    # assert os.path.isfile(os.path.join(processed_mmspro, 'fake.mmspro.idx'))
    assert os.path.isfile(os.path.join(processed_mmspro, 'fake.mmspro_h.index'))


# def test_down_and_process_camper(tmpdir, logger):
#     output_dir = tmpdir.mkdir('process_camper')
#     temporary = os.path.join(output_dir, 'temp')
#     os.mkdir(temporary)
#     loc= download_camper_tar_gz(temporary, logger)
#     process_camper(loc, output_dir, logger)




"""
One test to stop so much pain, did you accidentally commit a test config?
This tests that the config is clean. It will fail when you are testing a working copy but not when you push to github.

"""

import pytest
import logging
from os import path
from mag_annotator.database_handler import DatabaseHandler
from mag_annotator.utils import setup_logger



def test_clean_config():
    empty_config = {
        "search_databases": {
        "kegg": None,
        "kofam_hmm": None,
        "kofam_ko_list": None,
        "uniref": None,
        "pfam": None,
        "dbcan": None,
        "viral": None,
        "peptidase": None,
        "vogdb": None
      },
      "custom_dbs": None,
      "database_descriptions": {
        "pfam_hmm_dat": None,
        "dbcan_fam_activities": None,
        "vog_annotations": None
      },
      "dram_sheets": {
        "genome_summary_form": None,
        "module_step_form": None,
        "etc_module_database": None,
        "function_heatmap_form": None,
        "amg_database": None
      },
      "description_db": None,
      "dram_version": None
    }
    logger = logging.getLogger('test_log')
    db_handler = DatabaseHandler(logger)
    assert db_handler.config == empty_config



"""
Defines the classes that are used to acsess sql data with sqlalchemy. Each data set that gets its descriptions stored in sql needs its database stored there also.
"""
from sqlalchemy import Column, String, create_engine
from sqlalchemy.ext.declarative import declarative_base

# TODO: Do all processing of descriptions here
# TODO: set up init statements that can parse the line into desired parameters

Base = declarative_base()

KEGG_DESCRIPTION_TABLE_NAME = 'kegg_description'


class KeggDescription(Base):
    __tablename__ = KEGG_DESCRIPTION_TABLE_NAME

    id = Column(String(20), primary_key=True, nullable=False, index=True)

    description = Column(String(100000))

    @property
    def serialize(self):
        return {
            'kegg_id': self.id,
            'kegg_description': self.description,
        }


UNIREF_DESCRIPTION_TABLE_NAME = 'uniref_description'


class UniRefDescription(Base):
    __tablename__ = UNIREF_DESCRIPTION_TABLE_NAME

    id = Column(String(40), primary_key=True, nullable=False, index=True)

    description = Column(String(1000))

    @property
    def serialize(self):
        return {
            'kegg_id': self.id,
            'kegg_description': self.description,
        }


PFAM_DESCRIPTION_TABLE_NAME = 'pfam_description'


class PfamDescription(Base):
    __tablename__ = PFAM_DESCRIPTION_TABLE_NAME

    id = Column(String(12), primary_key=True, nullable=False, index=True)

    description = Column(String(1000))

    @property
    def serialize(self):
        return {
            'pfam_id': self.id,
            'pfam_description': self.description,
        }


DBCAN_DESCRIPTION_TABLE_NAME = 'dbcan_description'


class DbcanDescription(Base):
    __tablename__ = DBCAN_DESCRIPTION_TABLE_NAME

    id = Column(String(30), primary_key=True, nullable=False, index=True)

    description = Column(String(1000))
    ec = Column(String(1000))

    @property
    def serialize(self):
        return {
            'dbcan_id': self.id,
            'dbcan_description': self.description,
            'dbcan_subfam_ec': self.ec,
        }

VIRAL_DESCRIPTION_TABLE_NAME = 'viral_description'


class ViralDescription(Base):
    __tablename__ = VIRAL_DESCRIPTION_TABLE_NAME

    id = Column(String(14), primary_key=True, nullable=False, index=True)

    description = Column(String(1000))

    @property
    def serialize(self):
        return {
            'viral_id': self.id,
            'viral_description': self.description,
        }


PEPTIDASE_DESCRIPTION_TABLE_NAME = 'peptidase_description'


class PeptidaseDescription(Base):
    __tablename__ = PEPTIDASE_DESCRIPTION_TABLE_NAME

    id = Column(String(10), primary_key=True, nullable=False, index=True)

    description = Column(String(1000))

    @property
    def serialize(self):
        return {
            'peptidase_id': self.id,
            'peptidase_description': self.description,
        }


VOGDB_DESCRIPTION_TABLE_NAME = 'vogdb_description'


class VOGDBDescription(Base):
    __tablename__ = VOGDB_DESCRIPTION_TABLE_NAME

    id = Column(String(10), primary_key=True, nullable=False, index=True)

    description = Column(String(1000))

    @property
    def serialize(self):
        return {
            'vogdb_id': self.id,
            'vogdb_description': self.description,
        }


def create_description_db(db_loc):
    engine = create_engine('sqlite:///%s' % db_loc)
    Base.metadata.create_all(engine)


TABLE_NAME_TO_CLASS_DICT = {KEGG_DESCRIPTION_TABLE_NAME: KeggDescription,
                            UNIREF_DESCRIPTION_TABLE_NAME: UniRefDescription,
                            PFAM_DESCRIPTION_TABLE_NAME: PfamDescription,
                            DBCAN_DESCRIPTION_TABLE_NAME: DbcanDescription,
                            VIRAL_DESCRIPTION_TABLE_NAME: ViralDescription,
                            PEPTIDASE_DESCRIPTION_TABLE_NAME: PeptidaseDescription,
                            VOGDB_DESCRIPTION_TABLE_NAME: VOGDBDescription}



"""DRAM-v distillation functions"""
import re
from os import path, mkdir
from functools import partial
from collections import defaultdict, Counter
from datetime import datetime
import warnings
import logging

import pandas as pd
import altair as alt

from mag_annotator.database_handler import DatabaseHandler
from mag_annotator.utils import setup_logger
from mag_annotator.summarize_genomes import get_ids_from_annotations_by_row, \
    get_ids_from_annotations_all, get_ordered_uniques, check_columns

VOGDB_TYPE_NAMES = {'Xr': 'Viral replication genes', 'Xs': 'Viral structure genes',
                    'Xh': 'Viral genes with host benefits', 'Xp': 'Viral genes with viral benefits',
                    'Xu': 'Viral genes with unknown function', 'Xx': 'Viral hypothetical genes'}
VIRUS_STATS_COLUMNS = ['VIRSorter category', 'Circular', 'Prophage', 'Gene count', 'Strand switches',
                       'potential AMG count', 'Transposase present', 'Possible Non-Viral Contig']
VIRAL_DISTILLATE_COLUMNS = ['gene', 'scaffold', 'gene_id', 'gene_description', 'category', 'header',
                            'subheader', 'module', 'auxiliary_score', 'amg_flags']
VIRAL_LIQUOR_HEADERS = ['Category', 'Function', 'AMG Genes', 'Genes Present', 'Contig Name', 'Present in Contig']
HEATMAP_CELL_HEIGHT = 10
HEATMAP_CELL_WIDTH = 10

defaultdict_list = partial(defaultdict, list)


def add_custom_ms(annotations, distillate_form):
    metabolic_genes = set(distillate_form.index)

    new_amg_flags = list()
    for gene, row in annotations.iterrows():
        if 'M' in row['amg_flags']:
            new_amg_flags.append(row['amg_flags'])
        else:
            gene_annotations = set(get_ids_from_annotations_all(pd.DataFrame(row).transpose()).keys())
            if len(metabolic_genes & gene_annotations) > 0:
                new_amg_flags.append(row['amg_flags'] + 'M')
            else:
                new_amg_flags.append(row['amg_flags'])
    return new_amg_flags


def filter_to_amgs(annotations, max_aux=4, remove_transposons=True, remove_fs=False):
    amgs = annotations[((annotations['amg_flags'].str.contains('M')) &
                          (annotations['amg_flags'].str.contains('V') == False) &
                          (annotations['amg_flags'].str.contains('A') == False) &
                          (annotations['amg_flags'].str.contains('P') == False) &
                          (annotations['auxiliary_score'] <= max_aux)
                           )]
    if remove_transposons:
        amgs = amgs[(amgs['amg_flags'].str.contains('T') == False)]
    if remove_fs:
        amgs = amgs[(amgs['amg_flags'].str.contains('F') == False)]
    return amgs


def get_strand_switches(strandedness):
    switches = 0
    strand = strandedness[0]
    for i in range(len(strandedness)):
        if strandedness[i] != strand:
            switches += 1
            strand = strandedness[i]
    return switches


def make_viral_stats_table(annotations, potential_amgs, groupby_column='scaffold'):
    amg_counts = potential_amgs.groupby(groupby_column).size()
    viral_stats_series = list()
    for scaffold, frame in annotations.groupby(groupby_column):
        # get virus information
        virus_categories = re.findall(r'-cat_\d$', scaffold)
        if len(virus_categories) > 0:
            virus_category = int(virus_categories[0].split('_')[-1])  # viral category
            virus_prophage = virus_category in [4, 5]  # virus is prophage
        else:
            virus_category = None
            virus_prophage = None
        virus_circular = len(re.findall(r'-circular-cat_\d$', scaffold)) == 1  # virus is circular
        virus_num_genes = len(frame)  # number of genes on viral contig
        virus_strand_switches = get_strand_switches(frame.strandedness)  # number of strand switches
        if scaffold in amg_counts:
            virus_number_amgs = amg_counts[scaffold]  # number of potential amgs
        else:
            virus_number_amgs = 0
        virus_transposase_present = sum(frame.is_transposon) > 0  # transposase on contig
        # virus_j_present = sum(['J' in i if not pd.isna(i) else False for i in frame.amg_flags]) > 0
        virus_j_present = sum([i == 'Xh' if not pd.isna(i) else False
                               for i in frame['vogdb_categories']]) / frame.shape[0]
        virus_data = pd.Series([virus_category, virus_circular, virus_prophage, virus_num_genes, virus_strand_switches,
                                virus_number_amgs, virus_transposase_present, virus_j_present],
                               index=VIRUS_STATS_COLUMNS, name=scaffold)
        # get vogdb categories
        # when vogdb has multiple categories only the first is taken
        gene_counts = Counter([i.split(';')[0] for i in frame.vogdb_categories.replace('', 'Xx')])
        named_gene_counts = {VOGDB_TYPE_NAMES[key]: value for key, value in gene_counts.items()}
        gene_counts_series = pd.Series(named_gene_counts, name=scaffold)
        viral_stats_series.append(pd.concat([virus_data, gene_counts_series]))
    return pd.DataFrame(viral_stats_series).fillna(0)


def make_viral_distillate(potential_amgs, genome_summary_form, amg_database, logger):
    """Make a summary of what in our database makes something a AMG or likly AMG to dram"""
    # Transform the amg database to make it more workable
    def look_up_metabolic_info(search_db, match_db, match_db_name):
        id_genes = set(match_db.index)
        return (
            (search_db
             .assign(gene_id = lambda x: x['ids'].apply(lambda y: y & id_genes))
             )[['gene_id', 'scaffold', 'auxiliary_score', 'amg_flags']]
            .explode('gene_id')
            .dropna(subset=['gene_id'])
            .merge(match_db, how='left', left_on='gene_id', right_index=True)
            .assign(gene_id_origin=match_db_name))

    amg_database_frame = (amg_database
                          .melt(value_vars=['KO', 'EC', 'PFAM'],
                                id_vars=['gene', 'module', 'metabolism',
                                         'reference', 'verified'],
                                value_name='gene_id')
                          .drop('variable', axis=1)
                          .assign(
                              gene_id=lambda x: x['gene_id'].apply(
                                  lambda y: [i.strip() for i in str(y).split(';')]))
                          .explode('gene_id')
                          .dropna(subset='gene_id')
                          .set_index('gene_id')
                          .rename(columns = {'gene': 'gene_description'})
                          )
    potential_amgs = potential_amgs.assign(ids=get_ids_from_annotations_by_row(potential_amgs))
    metabolic_df = look_up_metabolic_info(potential_amgs, genome_summary_form, 'genome_summary_form')
    amg_df = look_up_metabolic_info(potential_amgs, amg_database_frame, 'amg_database')
    missing = list(set(potential_amgs.index) - (set(metabolic_df.index) | (set(amg_df.index)) ))
    # evaluate what is mising
    logger.warning(f"No distillate information found for {len(missing)} genes.")
    logger.debug('\n'.join(missing))


    summary = pd.concat([
        metabolic_df,
        amg_df,
        potential_amgs.loc[missing, ['scaffold', 'auxiliary_score', 'amg_flags']]])
    summary.reset_index(inplace=True, drop=False, names='gene')
    return summary


def make_vgf_order(amgs):
    amg_score_dict = {scaffold: ((1/frame['auxiliary_score']).sum(), len(frame))
                      for scaffold, frame in amgs.groupby('scaffold')}
    amg_scores = pd.DataFrame.from_dict(amg_score_dict, columns=['AMG_score', 'AMG_count'],
                                        orient='index')
    return list(amg_scores.sort_values(['AMG_score', 'AMG_count'], ascending=False).index)


def make_amg_count_column(potential_amgs, vgf_order=None):
    # build count column
    amg_counts = pd.DataFrame(Counter(potential_amgs.scaffold).items(), columns=['Contig Name', 'Number'])
    amg_counts['AMG Count'] = 'AMG Count'
    text = alt.Chart(amg_counts, width=HEATMAP_CELL_WIDTH+10, height=HEATMAP_CELL_HEIGHT*len(amg_counts)).encode(
                     x=alt.X('AMG Count', title=None, axis=alt.Axis(labelLimit=0, labelAngle=90)),
                     y=alt.Y('Contig Name', title=None, axis=alt.Axis(labelLimit=0), sort=vgf_order),
                     text='Number'
                    ).mark_text()
    return text


def make_viral_functional_df(annotations, genome_summary_form, groupby_column='scaffold'):
    # build dict of ids per genome
    vgf_to_id_dict = defaultdict(defaultdict_list)
    for vgf, frame in annotations.groupby(groupby_column, sort=False):
        for gene, id_list in get_ids_from_annotations_by_row(frame).items():
            for id_ in id_list:
                vgf_to_id_dict[vgf][id_].append(gene)
    # build long from data frame
    rows = list()
    for category, category_frame in genome_summary_form.groupby('sheet'):
        for header, header_frame in category_frame.groupby('module'):
            header_id_set = set(header_frame.index.to_list())
            curr_rows = list()
            for vgf, id_dict in vgf_to_id_dict.items():
                present_in_bin = False
                functions_present = list()
                amgs_present = list()
                for id_, amgs in id_dict.items():
                    if id_ in header_id_set:
                        present_in_bin = True
                        functions_present.append(id_)
                        amgs_present += amgs
                curr_rows.append([category, header, ', '.join(amgs_present), ', '.join(functions_present), vgf,
                                  present_in_bin])
            if sum([i[-1] for i in curr_rows]) > 0:
                rows += curr_rows
    return pd.DataFrame(rows, columns=VIRAL_LIQUOR_HEADERS)


def make_viral_functional_heatmap(functional_df, vgf_order=None):
    # build heatmaps
    charts = list()
    for i, (group, frame) in enumerate(functional_df.groupby('Category', sort=False)):
        # set variables for chart
        function_order = get_ordered_uniques(list(frame['Function']))
        num_vgfs_in_frame = len(set(frame['Contig Name']))
        chart_width = HEATMAP_CELL_WIDTH * len(function_order)
        chart_height = HEATMAP_CELL_HEIGHT * num_vgfs_in_frame
        # set up colors for chart
        rect_colors = alt.Color('Present in Contig',
                                legend=alt.Legend(symbolType='square', values=[True, False]),
                                sort=[True, False],
                                scale=alt.Scale(range=['#e5f5f9', '#2ca25f']))
        # define chart
        # TODO: Figure out how to angle title to take up less space
        c = alt.Chart(frame, title=alt.TitleParams(group)).encode(
            x=alt.X('Function', title=None, axis=alt.Axis(labelLimit=0, labelAngle=90), sort=function_order),
            y=alt.Y('Contig Name', axis=alt.Axis(title=None, labels=False, ticks=False), sort=vgf_order),
            tooltip=[alt.Tooltip('Contig Name'),
                     alt.Tooltip('Category'),
                     alt.Tooltip('Function'),
                     alt.Tooltip('AMG Genes'),
                     alt.Tooltip('Genes Present')]
        ).mark_rect().encode(color=rect_colors).properties(
            width=chart_width,
            height=chart_height)
        charts.append(c)
    # merge and return
    function_heatmap = alt.hconcat(*charts, spacing=5)
    return function_heatmap


def summarize_vgfs(input_file, output_dir, groupby_column='scaffold', max_auxiliary_score=3,
                   remove_transposons=False, remove_fs=False, custom_distillate=None,
                   log_file_path:str=None, config_loc=None):
    # make output folder
    mkdir(output_dir)
    if log_file_path is None:
        log_file_path = path.join(output_dir, "distill.log")
    logger = logging.getLogger('distillation_log')
    setup_logger(logger, log_file_path)
    logger.info(f"The log file is created at {log_file_path}")


    # set up
    annotations = pd.read_csv(input_file, sep='\t', index_col=0).fillna('')
    database_handler = DatabaseHandler(logger, config_loc=config_loc)
    if database_handler.config["dram_sheets"].get('genome_summary_form') is None:
        raise ValueError('Genome summary form location must be set in order to summarize genomes')
    genome_summary_form = pd.read_csv(database_handler.config['dram_sheets']['genome_summary_form'], sep='\t', index_col=0)
    if custom_distillate is not None:
        custom_distillate_form = pd.read_csv(custom_distillate, sep='\t', index_col=0)
        genome_summary_form = pd.concat([genome_summary_form, custom_distillate_form])
        # add M's from custom distillate
        annotations['amg_flags'] = add_custom_ms(annotations, custom_distillate_form)
    logger.info('Retrieved database locations and descriptions')

    # get potential AMGs
    potential_amgs = filter_to_amgs(annotations, max_aux=max_auxiliary_score,
                                    remove_transposons=remove_transposons, remove_fs=remove_fs)
    check_columns(potential_amgs, logger)
    logger.info('Determined potential amgs')

    # make distillate
    viral_genome_stats = make_viral_stats_table(annotations, potential_amgs, groupby_column)
    viral_genome_stats.to_csv(path.join(output_dir, 'vMAG_stats.tsv'), sep='\t')
    logger.info('Calculated viral genome statistics')

    viral_distillate = make_viral_distillate(
        potential_amgs,
        genome_summary_form,
        pd.read_csv(database_handler.config["dram_sheets"].get('amg_database'), sep='\t'),
        logger)
    viral_distillate.to_csv(path.join(output_dir, 'amg_summary.tsv'), sep='\t', index=None)
    logger.info('Generated AMG summary')

    # make liquor
    vgf_order = make_vgf_order(potential_amgs)
    amg_column = make_amg_count_column(potential_amgs, vgf_order)
    viral_function_df = make_viral_functional_df(potential_amgs, genome_summary_form, groupby_column=groupby_column)
    viral_functional_heatmap = make_viral_functional_heatmap(viral_function_df, vgf_order)
    product = alt.hconcat(amg_column, viral_functional_heatmap, spacing=5)
    product.save(path.join(output_dir, 'product.html'))
    logger.info('Generated product heatmap')
    logger.info("Completed distillation")



from os import path, remove, getenv
from pkg_resources import resource_filename
import json
import gzip
import logging
from shutil import copy2
import warnings
from datetime import datetime
from functools import partial
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

import pandas as pd

from mag_annotator import __version__ as current_dram_version
from mag_annotator.database_setup import TABLE_NAME_TO_CLASS_DICT, create_description_db
from mag_annotator.utils import divide_chunks, setup_logger

SEARCH_DATABASES = {
    "kegg",
    "kofam_hmm",
    "kofam_ko_list",
    "uniref",
    "pfam",
    "dbcan",
    "viral",
    "peptidase",
    "vogdb",
}
DRAM_SHEETS = (
    "genome_summary_form",
    "module_step_form",
    "etc_module_database",
    "function_heatmap_form",
    "amg_database",
)
DATABASE_DESCRIPTIONS = ("pfam_hmm", "dbcan_fam_activities", "vog_annotations")

# TODO: store all sequence db locs within database handler class
# TODO: store scoring information here e.g. bitscore_threshold, hmm cutoffs
# TODO: set up custom databases here
# TODO: in advanced config separate search databases, search database description files, description db, DRAM sheets
# TODO: ko_list should be parsed into the DB and stored as a database description file and not a search database


def get_config_loc():
    loc = getenv("DRAM_CONFIG_LOCATION")
    if loc:
        return loc
    else:
        return path.abspath(resource_filename("mag_annotator", "CONFIG"))


def clear_dict(val):
    if isinstance(val, dict):
        return {k: clear_dict(v) for k, v in val.items()}
    else:
        return None


class DatabaseHandler:
    def __init__(self, logger, config_loc=None):
        # read in new configuration
        # TODO: validate config file after reading it in
        if logger is None:
            logger = logging.getLogger("database_handler.log")
            # log_path = self.get_log_path()
            # setup_logger(logger, log_path)
            setup_logger(logger)
            logger.info(f"Logging to console")
        self.logger = logger
        if config_loc is None:
            config_loc = get_config_loc()
        self.load_config(config_loc)

        self.config_loc = config_loc

    def load_config(self, config_file):
        conf = json.loads(open(config_file).read())
        if len(conf) == 0:
            self.logger.warn("There is no config information in the provided file")
            self.clear_config(write_config=False)
        if "dram_version" not in conf:
            warnings.warn(
                "The DRAM version in your config is empty."
                " This may not be a problem, but if this"
                " import fails then you should check that"
                " the origin of the file is valid."
            )
            self.__construct_from_dram_pre_1_4_0(conf)
        else:
            conf_version = conf.get("dram_version")
            if conf_version is None:
                self.__construct_from_dram_pre_1_4_0(conf)
            elif conf_version not in {
                current_dram_version,
                "1.4.0",
                "1.4.0rc1",
                "1.4.0rc2",
                "1.4.0rc3",
                "1.4.0rc4",
            }:  # Known suported versions
                warnings.warn(
                    "The DRAM version in your config is not listed in the versions "
                    "that are known to work. This may not be a problem, but if this "
                    "import fails then you should contact suport."
                )
            self.__construct_default(conf)

    def get_log_path(self):
        path = self.config.get("log_path")
        if path is None:
            path = path.join(self.config_loc, "database_processing.log")
        return path

    def __construct_default(self, conf: dict):
        self.config = conf

        # set up description database connection
        description_loc = self.config.get("description_db")
        if description_loc is None:
            self.session = None
            warnings.warn("Database does not exist at path %s" % description_loc)
        elif not path.exists(description_loc):
            self.session = None
            warnings.warn("Database does not exist at path %s" % description_loc)
        else:
            self.start_db_session()

    def __construct_from_dram_pre_1_4_0(self, config_old):
        """
        Import older dram configs that predate 1.3


        :param config_old: A config with no dram version so older than 1.3
        """

        system_config_loc = get_config_loc()
        self.config = clear_dict(json.loads(open(system_config_loc).read()))
        self.config_loc = system_config_loc

        # read in configuration # TODO: validate config file after reading it in
        if 'viral_refseq' in config_old:
            config_old['viral'] = config_old.get('viral_refseq')
        if 'kofam' in config_old:
            config_old['kofam_hmm'] = config_old.get('kofam')
        if 'pfam_hmm_dat' in config_old:
            config_old['pfam_hmm'] = config_old.get('pfam_hmm_dat')
        self.config["search_databases"] = {
            key: value for key, value in config_old.items() if key in SEARCH_DATABASES
        }
        self.config["database_descriptions"] = {
            key: value
            for key, value in config_old.items()
            if key in DATABASE_DESCRIPTIONS
        }
        self.config["dram_sheets"] = {
            key: value for key, value in config_old.items() if key in DRAM_SHEETS
        }
        self.config["dram_version"] = current_dram_version

        # set up description database connection
        self.config["description_db"] = config_old.get("description_db")
        if self.config.get("description_db") is None:
            self.session = None
            warnings.warn(
                "Database does not exist at path %s" % self.config.get("description_db")
            )
        elif not path.exists(self.config.get("description_db")):
            self.session = None
            warnings.warn(
                "Database does not exist at path %s" % self.config.get("description_db")
            )
        else:
            self.start_db_session()

    def start_db_session(self):
        engine = create_engine("sqlite:///%s" % self.config.get("description_db"))
        db_session = sessionmaker(bind=engine)
        self.session = db_session()

    # functions for adding descriptions to tables
    def add_descriptions_to_database(self, description_list, db_name, clear_table=True):
        description_class = TABLE_NAME_TO_CLASS_DICT[db_name]
        if clear_table:
            self.session.query(description_class).delete()
        # TODO: try batching
        self.session.bulk_save_objects(
            [description_class(**i) for i in description_list]
        )
        self.session.commit()
        self.session.expunge_all()

    # functions for getting descriptions from tables
    def get_description(self, annotation_id, db_name, return_ob=False):
        return (
            self.session.query(TABLE_NAME_TO_CLASS_DICT[db_name])
            .filter_by(id=annotation_id)
            .one()
            .description
        )

    def get_descriptions(self, ids, db_name, description_name="description"):
        description_class = TABLE_NAME_TO_CLASS_DICT[db_name]
        descriptions = [
            des
            for chunk in divide_chunks(list(ids), 499)
            for des in self.session.query(description_class)
            .filter(description_class.id.in_(chunk))
            .all()
        ]
        # [des for des in self.session.query(description_class).filter(description_class.id.in_(list(ids))).all() ]
        # [i.id for i in self.session.query(TABLE_NAME_TO_CLASS_DICT['dbcan_description']).all()]
        if len(descriptions) == 0:
            warnings.warn(
                "No descriptions were found for your id's. Does this %s look like an id from %s"
                % (list(ids)[0], db_name)
            )
        return {i.id: i.__dict__[description_name] for i in descriptions}

    @staticmethod
    def get_database_names():
        return TABLE_NAME_TO_CLASS_DICT.keys()

    def get_settings_str(self):
        out_str = ""
        settings = self.config.get("setup_info")
        if settings is None:
            warnings.warn(
                "there are no settings, the config is corrupted or too old.",
                DeprecationWarning,
            )
            return "there are no settings, the config is corrupted or too old."
        for i in ["search_databases", "database_descriptions", "dram_sheets"]:
            out_str += "\n"
            for k in self.config.get(i):
                if settings.get(k) is not None:
                    out_str += f"\n{settings[k]['name']}:"
                    for l, w in settings[k].items():
                        if l == "name":
                            continue
                        out_str += f"\n    {l.title()}: {w}"
        return out_str

    def set_database_paths(
        self,
        kegg_loc=None,
        kofam_hmm_loc=None,
        kofam_ko_list_loc=None,
        uniref_loc=None,
        pfam_loc=None,
        pfam_hmm_loc=None,
        dbcan_loc=None,
        dbcan_fam_activities_loc=None,
        dbcan_subfam_ec_loc=None,
        viral_loc=None,
        peptidase_loc=None,
        vogdb_loc=None,
        vog_annotations_loc=None,
        description_db_loc=None,
        log_path_loc=None,
        genome_summary_form_loc=None,
        module_step_form_loc=None,
        etc_module_database_loc=None,
        function_heatmap_form_loc=None,
        amg_database_loc=None,
        write_config=True,
        update_description_db=False
    ):
        def check_exists_and_add_to_location_dict(loc, old_value):
            if loc is None:  # if location is none then return the old value
                return old_value
            if path.isfile(loc):  # if location exists return full path
                return path.realpath(loc)
            else:  # if the location doesn't exist then raise error
                raise ValueError("Database location does not exist: %s" % loc)

        locs = {
            "search_databases": {
                "kegg": kegg_loc,
                "kofam_hmm": kofam_hmm_loc,
                "kofam_ko_list": kofam_ko_list_loc,
                "uniref": uniref_loc,
                "pfam": pfam_loc,
                "dbcan": dbcan_loc,
                "viral": viral_loc,
                "peptidase": peptidase_loc,
                "vogdb": vogdb_loc,
            },
            "database_descriptions": {
                "pfam_hmm": pfam_hmm_loc,
                "dbcan_fam_activities": dbcan_fam_activities_loc,
                "dbcan_subfam_ec": dbcan_subfam_ec_loc,
                "vog_annotations": vog_annotations_loc,
            },
            "dram_sheets": {
                "genome_summary_form": genome_summary_form_loc,
                "module_step_form": module_step_form_loc,
                "etc_module_database": etc_module_database_loc,
                "function_heatmap_form": function_heatmap_form_loc,
                "amg_database": amg_database_loc,
            },
        }

        self.config.update(
            {
                i: {
                    k: check_exists_and_add_to_location_dict(
                        locs[i][k], self.config.get(i).get(k)
                    )
                    for k in locs[i]
                }
                for i in locs
            }
        )
        if update_description_db:
            self.populate_description_db(output_loc=description_db_loc)
        else:
            self.config["description_db"] = check_exists_and_add_to_location_dict(
                description_db_loc, self.config.get("description_db")
            )
        self.config["log_path"] = check_exists_and_add_to_location_dict(
            log_path_loc, self.config.get("log_path_db")
        )
        self.start_db_session()

        if write_config:
            self.write_config()


    def write_config(self, config_loc=None):
        if config_loc is None:
            config_loc = self.config_loc
        with open(config_loc, "w") as f:
            f.write(json.dumps(self.config, indent=2))

    @staticmethod
    def make_header_dict_from_mmseqs_db(mmseqs_db):
        mmseqs_headers_handle = open("%s_h" % mmseqs_db, "rb")
        mmseqs_headers = mmseqs_headers_handle.read().decode(errors="ignore")
        mmseqs_headers = [
            i.strip() for i in mmseqs_headers.strip().split("\n\x00") if len(i) > 0
        ]
        mmseqs_headers_split = []
        mmseqs_ids_unique = set()
        mmseqs_ids_not_unique = set()
        # TODO this could be faster with numpy
        for i in mmseqs_headers:
            header = {"id": i.split(" ")[0], "description": i}
            if header["id"] not in mmseqs_ids_unique:
                mmseqs_headers_split += [header]
                mmseqs_ids_unique.add(header["id"])
            else:
                mmseqs_ids_not_unique.add(header["id"])
        if len(mmseqs_ids_not_unique) > 0:
            warnings.warn(
                f"There are {len(mmseqs_ids_not_unique)} non unique headers "
                f"in {mmseqs_db}! You should definitly investigate this!"
            )
        return mmseqs_headers_split

    @staticmethod
    def process_pfam_descriptions(pfam_hmm):
        if pfam_hmm.endswith(".gz"):
            f = gzip.open(pfam_hmm, "r").read().decode("utf-8")
        else:
            f = open(pfam_hmm).read()
        entries = f.strip().split("//")
        description_list = list()
        for i, entry in enumerate(entries):
            if len(entry) > 0:
                entry = entry.split("\n")
                ascession = None
                description = None
                for line in entry:
                    line = line.strip()
                    if line.startswith("#=GF AC"):
                        ascession = line.split("   ")[-1]
                    if line.startswith("#=GF DE"):
                        description = line.split("   ")[-1]
                description_list.append({"id": ascession, "description": description})
        return description_list

    @staticmethod
    def process_dbcan_descriptions(dbcan_fam_activities, dbcan_subfam_ec):
        def line_reader(line):
            if not line.startswith("#") and len(line.strip()) != 0:
                line = line.strip().split()
                if len(line) == 1:
                    description = line[0]
                elif line[0] == line[1]:
                    description = " ".join(line[1:])
                else:
                    description = " ".join(line)
                return pd.DataFrame(
                    {"id": line[0], "description": description.replace("\n", " ")},
                    index=[0],
                )

        with open(dbcan_fam_activities) as f:
            description_data = pd.concat([line_reader(line) for line in f.readlines()])

        ec_data = pd.read_csv(
            dbcan_subfam_ec, sep="\t", names=["id", "id2", "ec"], comment="#"
        )[["id", "ec"]].drop_duplicates()
        ec_data = (
            pd.concat(
                [ec_data["id"], ec_data["ec"].str.split("|", expand=True)], axis=1
            )
            .melt(id_vars="id", value_name="ec")
            .dropna(subset=["ec"])[["id", "ec"]]
            .groupby("id")
            .apply(lambda x: ",".join(x["ec"].unique()))
        )
        ec_data = pd.DataFrame(ec_data, columns=["ec"]).reset_index()
        data = pd.merge(description_data, ec_data, how="outer", on="id").fillna("")
        return [i.to_dict() for _, i in data.iterrows()]

    @staticmethod
    def process_vogdb_descriptions(vog_annotations):
        annotations_table = pd.read_csv(vog_annotations, sep="\t", index_col=0)
        annotations_list = [
            {
                "id": vog,
                "description": "%s; %s"
                % (row["ConsensusFunctionalDescription"], row["FunctionalCategory"]),
            }
            for vog, row in annotations_table.iterrows()
        ]
        return annotations_list

    # TODO: Make option to build on description database that already exists?
    def populate_description_db(
        self, output_loc=None, select_db=None, update_config=True, erase_old_db=False
    ):
        if (
            self.config.get("description_db") is None and output_loc is None
        ):  # description db location must be set somewhere
            self.logger.critical(
                "Must provide output location if description db location is not set in configuration"
            )
            raise ValueError(
                "Must provide output location if description db location is not set in configuration"
            )
        if (
            output_loc is not None
        ):  # if new description db location is set then save it there
            self.config["description_db"] = output_loc
            self.start_db_session()
        # I don't think this is needed
        if path.exists(self.config.get("description_db")) and erase_old_db:
            remove(self.config.get("description_db"))
        create_description_db(self.config.get("description_db"))

        def check_db(db_name, db_function):
            # TODO add these sorts of checks to a separate function
            # if self.config.get('search_databases').get(db_name) is None:
            #     return
            # if not path.exists(self.config['search_databases'][db_name]):
            #     logger.warn(f"There is a path for the {db_name} db in the config, but there"
            #                 " is no file at that path. The path is:"
            #                 f"{self.config['search_databases'][db_name]}")
            #     return
            self.add_descriptions_to_database(
                db_function(), f"{db_name}_description", clear_table=True
            )
            self.config["setup_info"][db_name][
                "description_db_updated"
            ] = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
            self.logger.info(f"Description updated for the {db_name} database")

        # fill database
        mmseqs_database = ["kegg", "uniref", "viral", "peptidase"]
        process_functions = {
            i: partial(
                self.make_header_dict_from_mmseqs_db, self.config["search_databases"][i]
            )
            for i in mmseqs_database
            if self.config["search_databases"][i] is not None
        }
        # Use table names
        process_functions.update(
            {
                "pfam": partial(
                    self.process_pfam_descriptions,
                    self.config.get("database_descriptions")["pfam_hmm"],
                ),
                "dbcan": partial(
                    self.process_dbcan_descriptions,
                    self.config.get("database_descriptions")["dbcan_fam_activities"],
                    self.config.get("database_descriptions")["dbcan_subfam_ec"],
                ),
                "vogdb": partial(
                    self.process_vogdb_descriptions,
                    self.config.get("database_descriptions")["vog_annotations"],
                ),
            }
        )
        if select_db is not None:
            process_functions = {
                i: k for i, k in process_functions.items() if i in select_db
            }

        for i, k in process_functions.items():
            check_db(i, k)

        if update_config:  # if new description db is set then save it
            self.write_config()

    def filter_db_locs(
        self, low_mem_mode=False, use_uniref=True, use_vogdb=True, master_list=None
    ):
        if master_list is None:
            dbs_to_use = self.config["search_databases"].keys()
        else:
            dbs_to_use = master_list
        # filter out dbs for low mem mode
        if low_mem_mode:
            if ("kofam_hmm" not in self.config.get("search_databases")) or (
                "kofam_ko_list" not in self.config.get("search_databases")
            ):
                raise ValueError(
                    "To run in low memory mode KOfam must be configured for use in DRAM"
                )
            dbs_to_use = [i for i in dbs_to_use if i not in ("uniref", "kegg", "vogdb")]
        # check on uniref status
        if use_uniref:
            if "uniref" not in self.config.get("search_databases"):
                warnings.warn(
                    "Sequences will not be annoated against uniref as it is not configured for use in DRAM"
                )
        else:
            dbs_to_use = [i for i in dbs_to_use if i != "uniref"]
        # check on vogdb status
        if use_vogdb:
            if "vogdb" not in self.config.get("search_databases"):
                warnings.warn(
                    "Sequences will not be annoated against VOGDB as it is not configured for use in DRAM"
                )
        else:
            dbs_to_use = [i for i in dbs_to_use if i != "vogdb"]
        self.config["search_databases"] = {
            key: value
            for key, value in self.config.get("search_databases").items()
            if key in dbs_to_use
        }

    def clear_config(self, write_config=False):
        self.config = {
            "search_databases": {},
            "database_descriptions": {},
            "dram_sheets": {},
            "dram_version": current_dram_version,
            "description_db": None,
            "setup_info": {},
            "log_path": None,
        }
        if write_config:
            self.write_config()


def set_database_paths(clear_config=False, **kargs):
    # TODO Add tests
    db_handler = DatabaseHandler(None)
    if clear_config:
        db_handler.clear_config(write_config=True)
    db_handler.set_database_paths(**kargs, write_config=True)


def print_database_locations(config_loc=None):
    conf = DatabaseHandler(None, config_loc)
    # search databases
    print("Processed search databases")
    print("KEGG db: %s" % conf.config.get("search_databases").get("kegg"))
    print("KOfam db: %s" % conf.config.get("search_databases").get("kofam_hmm"))
    print(
        "KOfam KO list: %s" % conf.config.get("search_databases").get("kofam_ko_list")
    )
    print("UniRef db: %s" % conf.config.get("search_databases").get("uniref"))
    print("Pfam db: %s" % conf.config.get("search_databases").get("pfam"))
    print("dbCAN db: %s" % conf.config.get("search_databases").get("dbcan"))
    print("RefSeq Viral db: %s" % conf.config.get("search_databases").get("viral"))
    print(
        "MEROPS peptidase db: %s" % conf.config.get("search_databases").get("peptidase")
    )
    print("VOGDB db: %s" % conf.config.get("search_databases").get("vogdb"))
    # database descriptions used during description db population
    print()
    print("Descriptions of search database entries")
    print("Pfam hmm dat: %s" % conf.config.get("database_descriptions").get("pfam_hmm"))
    print(
        "dbCAN family activities: %s"
        % conf.config.get("database_descriptions").get("dbcan_fam_activities")
    )
    print(
        "VOG annotations: %s"
        % conf.config.get("database_descriptions").get("vog_annotations")
    )
    print()
    # description database
    print("Description db: %s" % conf.config.get("description_db"))
    print()
    # DRAM sheets
    print("DRAM distillation sheets")
    print(
        "Genome summary form: %s"
        % conf.config.get("dram_sheets").get("genome_summary_form")
    )
    print(
        "Module step form: %s" % conf.config.get("dram_sheets").get("module_step_form")
    )
    print(
        "ETC module database: %s"
        % conf.config.get("dram_sheets").get("etc_module_database")
    )
    print(
        "Function heatmap form: %s"
        % conf.config.get("dram_sheets").get("function_heatmap_form")
    )
    print("AMG database: %s" % conf.config.get("dram_sheets").get("amg_database"))


def print_database_settings(config_loc=None):
    conf = DatabaseHandler(None, config_loc)
    print(conf.get_settings_str())


def populate_description_db(output_loc=None, select_db=None, config_loc=None):
    db_handler = DatabaseHandler(None, config_loc)
    db_handler.populate_description_db(output_loc, select_db)


def export_config(output_file=None):
    config_loc = get_config_loc()
    if output_file is None:
        print(open(config_loc).read())
    else:
        copy2(config_loc, output_file)


def import_config(config_loc):
    system_config = get_config_loc()
    db_handler = DatabaseHandler(None, config_loc)
    with open(system_config, "w") as outfile:
        json.dump(db_handler.config, outfile, indent=2)
    print("Import, appears to be successfull.")


def mv_db_folder(new_location: str = "./", old_config_file: str = None):
    new_location = path.abspath(new_location)
    old_config_file = path.abspath(old_config_file)
    db_handler = DatabaseHandler(None)
    if old_config_file is not None:
        db_handler.load_config(old_config_file)
    paths = ["search_databases", "dram_sheets", "database_descriptions"]

    def auto_move_path(k: str, v: str):
        if v is None:
            db_handler.logger.warn(f"The path for {k} was not set, so can't update.")
            return
        new_path = path.join(new_location, path.basename(v))
        if not path.exists(new_path):
            db_handler.logger.warn(
                f"There is no file at path {new_path},"
                f" so no new location will be set for {k}."
            )
            return
        db_handler.logger.info(f"Moving {k} to {new_path}")
        db_handler.set_database_paths(**{f"{k}_loc": new_path}, write_config=True)

    auto_move_path("log_path", db_handler.config.get("log_path"))
    auto_move_path("description_db", db_handler.config.get("description_db"))
    for i in paths:
        for k, v in db_handler.config.get(i).items():
            auto_move_path(k, v)



"A simple script to check if there are groups of names in a set of fasta files"
from itertools import permutations
from collections import Counter

DFT_HEAD_CHAR = '>'


def fasta_dup_check(fasta:str, head_char=DFT_HEAD_CHAR) -> set:
    with open(fasta) as fa:
        headers = [line[1:].strip() for line in fa if line[0] == head_char]
    header_set = set(headers)
    if len(headers) >  len(header_set):
        raise ValueError(f"The FASTA file {fasta} contains duplicate headers,"
                         " you must correct this before continuing. The duplicate"
                         f" headers are: {[k for k, v in Counter(headers).items() if v > 1]}")
    return header_set


def __check_sets(i, j):
    if len(i[0]) + len(j[0]) > len(i[0].union(j[0])):
        dup_headers = [k for k, v in Counter(list(i[0]) + list(j[0])).items() if v > 1]
        if len(dup_headers) < 100:
            raise ValueError(f"The FASTA files {i[1]} and {j[1]} contains duplicate headers,"
                             " you must correct this before continuing. The duplicate"
                             f" headers are {dup_headers}")
        else:
            raise ValueError(f"The FASTA files {i[1]} and {j[1]} contains duplicate headers,"
                             " you must correct this before continuing. The first 100 duplicate"
                             f" headers are {dup_headers[:100]}")


def fastas_dup_check(fastas:list, head_char=DFT_HEAD_CHAR) -> bool:
    headers = [fasta_dup_check(fasta, head_char) for fasta in fastas]
    if sum([len(i) for i in headers]) > len(set().union(*headers)):
        for i in permutations(zip(headers, fastas), 2):
            __check_sets(i[0], i[1])
    return True
